# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import warnings
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Hashable,
    List,
    Optional,
    Protocol,
    Tuple,
    Union,
    overload,
)

import cachetools
import numpy
from pyproj.crs import CRS as _CRS
from pyproj.enums import WktVersion
from pyproj.exceptions import CRSError
from pyproj.transformer import Transformer

from .types import XY, Unset

EPSG_UNSET = 0


class CRSLike(Protocol):
    """CRS Like object."""

    def to_wkt(self, *args, **kw) -> str: ...


_crs_cache: Dict[Hashable, Tuple[_CRS, str, Optional[int]]] = {}


def _make_crs_key(crs_spec: Union[int, str, Hashable, CRSLike]) -> Hashable:
    if isinstance(crs_spec, str):
        normed_epsg = crs_spec.upper()
        if normed_epsg.startswith("EPSG:"):
            return normed_epsg
        return crs_spec
    if isinstance(crs_spec, int):
        return f"EPSG:{crs_spec}"
    if isinstance(crs_spec, Hashable):
        return crs_spec

    return crs_spec.to_wkt()


@cachetools.cached(_crs_cache, key=_make_crs_key)
def _make_crs(
    crs_spec: Union[str, int, _CRS, CRSLike]
) -> Tuple[_CRS, str, Optional[int]]:
    epsg = EPSG_UNSET
    if isinstance(crs_spec, str):
        crs = _CRS.from_user_input(crs_spec)
    elif isinstance(crs_spec, int):
        epsg = crs_spec
        crs = _CRS.from_epsg(crs_spec)
    elif isinstance(crs_spec, _CRS):
        crs = crs_spec
    else:
        crs = _CRS.from_wkt(crs_spec.to_wkt())

    crs_str = str(crs)
    crs_str_u = crs_str.upper()
    if crs_str_u.startswith("EPSG:"):
        crs_str = crs_str_u
        epsg = int(crs_str.split(":", 1)[1])

    return (crs, crs_str, epsg)


def _make_crs_transform_key(from_crs, to_crs, always_xy):
    return (id(from_crs), id(to_crs), always_xy)


@cachetools.cached({}, key=_make_crs_transform_key)
def _make_crs_transform(from_crs: _CRS, to_crs: _CRS, always_xy: bool) -> Transformer:
    return Transformer.from_crs(from_crs, to_crs, always_xy=always_xy)


class CRS:
    """
    Wrapper around :py:class:`pyproj.crs.CRS` for backwards compatibility.
    """

    DEFAULT_WKT_VERSION = WktVersion.WKT2_2019
    """Default version for WKT: WKT2_2019"""

    __slots__ = ("_crs", "_epsg", "_str")

    def __init__(self, crs_spec: Any):
        """
        Construct CRS object from *something*.

        :param crs_spec:
           String representation of a CRS, often an EPSG code like ``'EPSG:4326'``. Can also be any
           object that implements ``.to_epsg()`` or ``.to_wkt()``.

        :raises: :py:class:`pyproj.exceptions.CRSError`
        """

        if isinstance(crs_spec, (str, int, _CRS)):
            self._crs, self._str, self._epsg = _make_crs(crs_spec)
        elif isinstance(crs_spec, CRS):
            self._crs = crs_spec._crs
            self._epsg = crs_spec._epsg
            self._str = crs_spec._str
        elif isinstance(crs_spec, dict):
            self._crs, self._str, self._epsg = _make_crs(_CRS.from_dict(crs_spec))
        elif hasattr(crs_spec, "to_wkt"):
            self._crs, self._str, self._epsg = _make_crs(crs_spec)
        else:
            raise CRSError(f"Unexpected input encountered: {crs_spec}")

    def __getstate__(self):
        return {"crs_str": self._str}

    def __setstate__(self, state):
        self.__init__(state["crs_str"])

    def to_wkt(self, pretty: bool = False, version: Optional[WktVersion] = None) -> str:
        """
        Generate WKT representation of this CRS.

        :param pretty: If ``True`` generate multi-line WKT.
        :param version: Specify WKT version.
        """
        if version is None:
            version = self.DEFAULT_WKT_VERSION

        return self._crs.to_wkt(pretty=pretty, version=version)

    @property
    def wkt(self) -> str:
        """WKT representation of this CRS."""
        return self.to_wkt()

    def to_epsg(self) -> Optional[int]:
        """
        EPSG Code of the CRS or ``None``.
        """
        if self._epsg == EPSG_UNSET:
            self._epsg = self._crs.to_epsg()
        return self._epsg

    @property
    def epsg(self) -> Optional[int]:
        """
        EPSG Code of the CRS or ``None``.
        """
        return self.to_epsg()

    @property
    def semi_major_axis(self):
        """Semi-major axis of the ellipsoid."""
        return self._crs.ellipsoid.semi_major_metre

    @property
    def semi_minor_axis(self):
        """Semi-minor axis of the ellipsoid."""
        return self._crs.ellipsoid.semi_minor_metre

    @property
    def inverse_flattening(self):
        """Inverse flattening of the ellipsoid."""
        return self._crs.ellipsoid.inverse_flattening

    @property
    def geographic(self) -> bool:
        """True if CRS is geographic."""
        return self._crs.is_geographic

    @property
    def projected(self) -> bool:
        """True if CRS is projected."""
        return self._crs.is_projected

    @property
    def dimensions(self) -> Tuple[str, str]:
        """
        List of dimension names of the CRS.

        The ordering of the names is intended to reflect the :py:class:`numpy.ndarray` axis order of the
        loaded raster.
        """
        if self.geographic:
            return "latitude", "longitude"

        if self.projected:
            return "y", "x"

        raise ValueError("Neither projected nor geographic")  # pragma: no cover

    @property
    def units(self) -> Tuple[str, str]:
        """
        List of dimension units of the CRS.

        The ordering of the units is intended to reflect the :py:class:`numpy.ndarray` axis order of the
        loaded raster.
        """
        if self.geographic:
            return "degrees_north", "degrees_east"

        if self.projected:
            _dir_renames = {"north": "y", "south": "y", "east": "x", "west": "x"}
            units = {
                _dir_renames.get(ax.direction, ax.direction): ax.unit_name
                for ax in self._crs.axis_info
            }
            return units.get("y", ""), units.get("x", "")

        raise ValueError("Neither projected nor geographic")  # pragma: no cover

    @property
    def authority(self) -> Tuple[str, Union[str, int]]:
        """
        Get ``(authority_name, code)`` tuple.

        :returns: ``("", "")`` when not available
        """
        if self._epsg:
            return ("EPSG", self._epsg)

        if (r := self._crs.to_authority()) is not None:
            name, code = r
            try:
                return (name, int(code))
            except ValueError:  # pragma: nocover
                return (name, code)

        return ("", "")

    def __str__(self) -> str:
        return self._str

    def __hash__(self) -> int:
        return hash(self._crs)

    def __repr__(self) -> str:
        return f"CRS('{self._str}')"

    def __eq__(self, other) -> bool:
        if not isinstance(other, CRS):
            try:
                other = CRS(other)
            except Exception:  # pylint: disable=broad-except
                return False

        if self._crs is other._crs:
            return True

        if self._epsg and other._epsg:
            return self._epsg == other._epsg

        if self._str == other._str:
            return True

        return self._crs == other._crs

    def __ne__(self, other) -> bool:
        return not self == other

    @property
    def proj(self) -> _CRS:
        """Access :py:class:`pyproj.crs.CRS` object that this wraps."""
        return self._crs

    @property
    def valid_region(self) -> Optional["geom.Geometry"]:
        """
        Return valid region of this CRS.

        :returns: Bounding box in Lon/Lat as a 4 point Polygon in EPSG:4326.
        :returns: ``None`` if valid region is not defined for this CRS
        """

        from . import geom  # pylint: disable=import-outside-toplevel

        aou = self._crs.area_of_use
        if aou is None:
            return None
        return geom.box(aou.west, aou.south, aou.east, aou.north, "EPSG:4326")

    @property
    def crs_str(self) -> str:
        """DEPRECATED"""
        warnings.warn(
            "Please use `str(crs)` instead of `crs.crs_str`",
            category=DeprecationWarning,
        )
        return self._str

    def transformer_to_crs(
        self, other: "CRS", always_xy: bool = True
    ) -> Callable[[Any, Any], Tuple[Any, Any]]:
        """
        Build coordinate transformer to other projection.

        Returns a function that maps ``x, y -> x', y'`` where ``x, y`` are coordinates in this CRS,
        stored either as scalars or :py:class:`numpy.ndarray` objects, and ``x', y'`` are the same
        points in the ``other`` CRS.

        :param other:
              Destination CRS

        :param always_xy:
              If true, the transform method will accept as input and return as output coordinates
              using the traditional GIS order, that is longitude, latitude for geographic CRS and
              easting, northing for most projected CRS.
        """

        # pylint: disable=protected-access
        tr = _make_crs_transform(self._crs, other._crs, always_xy=always_xy)

        def result(x, y, **kw):
            rx, ry = tr.transform(x, y, **kw)  # pylint: disable=unpacking-non-sequence

            if not isinstance(rx, numpy.ndarray) or not isinstance(ry, numpy.ndarray):
                return (rx, ry)

            missing = numpy.isnan(rx) | numpy.isnan(ry)
            rx[missing] = numpy.nan
            ry[missing] = numpy.nan
            return (rx, ry)

        return result

    def __dask_tokenize__(self):
        return ("odc.geo.crs.CRS", str(self))

    @staticmethod
    def utm(
        x: Union[float, int, XY[float], "geom.Geometry", "geom.BoundingBox"],
        y: Optional[float] = None,
        /,
        datum_name: str = "WGS 84",
    ) -> "CRS":
        """
        Construct appropriate UTM CRS for a given point.

        Uses CRS database query methods from :py:mod:`pyproj` to locate
        appropriate UTM CRS.

        :params datum_name: The name of the datum in the CRS name ('NAD27', 'NAD83', 'WGS 84', ...)
        """
        # pylint: disable=import-outside-toplevel,no-name-in-module
        from pyproj.database import query_utm_crs_info

        from . import geom

        if isinstance(x, geom.BoundingBox):
            _bbox = x
        elif isinstance(x, geom.Geometry):
            if x.crs is not None:
                _bbox = x.to_crs("epsg:4326").boundingbox
            else:
                # assume already in lon/lat
                _bbox = x.boundingbox
        elif isinstance(x, (float, int)):
            if y is None:
                y = 0.0
            _bbox = geom.BoundingBox(x, y, x, y)
        else:
            x, y = x.xy
            _bbox = geom.BoundingBox(x, y, x, y)

        return _pick_best_crs(
            _bbox.polygon,
            [
                CRS(f"{info.auth_name}:{info.code}")
                for info in query_utm_crs_info(
                    datum_name=datum_name,
                    area_of_interest=_bbox.aoi,
                )
            ],
        )


class CRSMismatchError(ValueError):
    """
    CRS Mismatch Error.

    Raised when geometry operation is attempted on geometries in different
    coordinate references.
    """


SomeCRS = Union[str, int, CRS, _CRS, Dict[str, Any]]
MaybeCRS = Union[SomeCRS, Unset, None]

# fmt: off
@overload
def norm_crs(crs: SomeCRS) -> CRS: ...
@overload
def norm_crs(crs: SomeCRS, ctx: Any) -> CRS: ...
@overload
def norm_crs(crs: Union[None, Unset]) -> None: ...
@overload
def norm_crs(crs: Union[None, Unset], ctx: Any) -> None: ...
# fmt: on


def norm_crs(crs: MaybeCRS, ctx=None) -> Optional[CRS]:
    """Normalise CRS representation."""
    if isinstance(crs, CRS):
        return crs
    if crs is None or isinstance(crs, Unset):
        return None
    if isinstance(crs, str):
        _txt = crs.lower()
        if _txt.startswith("utm"):
            assert ctx is not None

            utm_crs = CRS.utm(ctx)
            if _txt == "utm":
                return utm_crs

            utm_zone = utm_crs.proj.utm_zone
            epsg = utm_crs.epsg
            assert utm_zone is not None
            assert epsg is not None
            if _txt == "utm-n" and utm_zone.endswith("S"):
                utm_crs = CRS(epsg - 100)
            elif _txt == "utm-s" and utm_zone.endswith("N"):
                utm_crs = CRS(epsg + 100)
            return utm_crs
    return CRS(crs)


def norm_crs_or_error(crs: MaybeCRS, ctx=None) -> CRS:
    """Normalise CRS representation, raise error if input is ``None``."""
    _crs = norm_crs(crs, ctx=ctx)
    if _crs is None:
        raise ValueError("Expect valid CRS")
    return _crs


def crs_units_per_degree(
    crs: SomeCRS,
    lon: Union[float, Tuple[float, float]],
    lat: float = 0,
    step: float = 0.1,
) -> float:
    """
    Helper method for converting resolution between meters/degrees.

    Compute number of CRS units per degree for a projected CRS at a given location
    in lon/lat.

    Location can be supplied as a tuple or as two arguments.

    :param crs: CRS
    :param lon: Either longitude or ``(lon, lat)`` tuple
    :param lat: Latitude or ignored if ``lon`` was a tuple
    :param step: Length of the segment in degrees used to estimate relative scale change

    :returns:  A floating number ``S`` such that ``S*degrees -> meters``
    """

    from . import geom  # pylint: disable=import-outside-toplevel

    if isinstance(lon, tuple):
        lon, lat = lon

    lon2 = lon + step
    if lon2 > 180:
        lon2 = lon - step

    ll = geom.line([(lon, lat), (lon2, lat)], "EPSG:4326")
    xy = ll.to_crs(crs)

    return xy.length / step


def _pick_best_crs(poly: "geom.Geometry", crs_candidates: List[CRS]) -> CRS:
    # pylint: disable=import-outside-toplevel
    from . import geom

    def overlap_pct(crs: CRS) -> float:
        crs_region = crs.valid_region
        if crs_region is None:
            return 1  # pragma: nocover
        return (crs_region & poly).area / poly.area

    if len(crs_candidates) < 1:
        raise ValueError("No candidate CRSs found")

    if len(crs_candidates) > 1 and poly.area > 1e-9:
        if poly.crs is None:
            poly = geom.Geometry(poly.geom, "epsg:4326")

        crs_candidates = sorted(crs_candidates, key=overlap_pct, reverse=True)

    return crs_candidates[0]


if TYPE_CHECKING:
    from . import geom  # pragma: no cover



import json
import lzma
import threading

from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, Optional, Tuple
from warnings import catch_warnings, filterwarnings

from ..crs import MaybeCRS, norm_crs
from ..geom import Geometry, box, multigeom


def data_path(fname: Optional[str] = None) -> Path:
    """Location of data folder or file within."""
    prefix = Path(__file__).parent
    if fname is None:
        return prefix
    return prefix / fname


@lru_cache
def ocean_geojson() -> Dict[str, Any]:
    with lzma.open(data_path("ocean.geojson.xz"), "rt") as xz:
        return json.load(xz)


def ocean_geom(
    crs: MaybeCRS = None, bbox: Optional[Tuple[float, float, float, float]] = None
) -> Geometry:
    """Return world oceans geometry."""
    gjson = ocean_geojson()
    gg = multigeom([Geometry(f["geometry"], "epsg:4326") for f in gjson["features"]])
    if bbox is not None:
        with catch_warnings():
            filterwarnings("ignore")
            gg = gg & box(*bbox, "epsg:4326")
    crs = norm_crs(crs)
    if crs is not None:
        gg = gg.to_crs(crs)

    return gg


@lru_cache
def gbox_css() -> str:
    with open(data_path("gbox.css"), "rt", encoding="utf8") as src:
        return src.read()


class _CachedGeoDataFrame:
    _instance = None

    # Override in sub-classes
    _lock = threading.Lock()
    _data_url = ""

    def __init__(self):
        # Thread safe class-cached dataload
        if self._instance is None:
            with self._lock:
                if self._instance is None:
                    self.__class__._instance = self._load_from_url()

    def _load_from_url(self):
        # pylint: disable=import-outside-toplevel
        import geopandas as gpd

        with catch_warnings():
            filterwarnings("ignore", category=FutureWarning)
            df = gpd.read_file(self._data_url)
        return df


class Countries(_CachedGeoDataFrame):
    """
    Cache-wrapper around the Natural Earth low-res countries geodataset.
    """

    _lock = threading.Lock()
    _data_url = (
        "https://naciscdn.org/naturalearth/110m/cultural/ne_110m_admin_0_countries.zip"
    )

    def frame_by_iso3(self, iso3):
        df = self._instance
        return df[df.ISO_A3 == iso3]


def country_geom(iso3: str, crs: MaybeCRS = None) -> Geometry:
    """
    Extract geometry for a country from geopandas sample data.
    """
    # pylint: disable=import-outside-toplevel
    from ..converters import from_geopandas

    countries = Countries()
    (gg,) = from_geopandas(countries.frame_by_iso3(iso3))
    crs = norm_crs(crs)
    if crs is not None:
        gg = gg.to_crs(crs)
    return gg



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import math
from random import uniform

import numpy as np
import pytest
from affine import Affine

from odc.geo import CRS, AnchorEnum, geom, res_, resyx_, wh_, xy_
from odc.geo.geobox import GeoBox, scaled_down_geobox
from odc.geo.gridspec import GridSpec
from odc.geo.math import affine_from_pts, decompose_rws, is_affine_st, stack_xy
from odc.geo.overlap import (
    LinearPointTransform,
    ReprojectInfo,
    _can_paste,
    compute_axis_overlap,
    compute_output_geobox,
    compute_reproject_roi,
    get_scale_at_point,
    native_pix_transform,
)
from odc.geo.roi import (
    roi_is_empty,
    roi_normalise,
    roi_shape,
    scaled_down_roi,
    scaled_up_roi,
)
from odc.geo.testutils import AlbersGS, epsg3577, epsg3857, epsg4326, mkA


def diff_affine(A: Affine, B: Affine) -> float:
    return math.sqrt(sum((a - b) ** 2 for a, b in zip(A, B)))


def test_affine_checks():
    assert is_affine_st(mkA(scale=(1, 2), translation=(3, -10))) is True
    assert is_affine_st(mkA(scale=(1, -2), translation=(-3, -10))) is True
    assert is_affine_st(mkA(rot=0.1)) is False
    assert is_affine_st(mkA(shear=0.4)) is False


def test_affine_rsw():
    def run_test(a, scale, shear=0, translation=(0, 0), tol=1e-8):
        A = mkA(a, scale=scale, shear=shear, translation=translation)

        R, W, S = decompose_rws(A)

        assert diff_affine(A, R * W * S) < tol
        assert diff_affine(S, mkA(0, scale)) < tol
        assert diff_affine(R, mkA(a, translation=translation)) < tol

    for a in (0, 12, 45, 33, 67, 89, 90, 120, 170):
        run_test(a, (1, 1))
        run_test(a, (0.5, 2))
        run_test(-a, (0.5, 2))

        run_test(a, (1, 2))
        run_test(-a, (1, 2))

        run_test(a, (2, -1))
        run_test(-a, (2, -1))

    run_test(0, (3, 4), 10)
    run_test(-33, (3, -1), 10, translation=(100, -333))


def test_fit():
    def run_test(A, n, tol=1e-5):
        X = [xy_(uniform(0, 1), uniform(0, 1)) for _ in range(n)]
        Y = [xy_(A * pt.xy) for pt in X]
        A_ = affine_from_pts(X, Y)

        assert diff_affine(A, A_) < tol

    A = mkA(13, scale=(3, 4), shear=3, translation=(100, -3000))

    run_test(A, 3)
    run_test(A, 10)

    run_test(mkA(), 3)
    run_test(mkA(), 10)


def test_scale_at_point():
    def mk_transform(sx, sy):
        A = mkA(37, scale=(sx, sy), translation=(2127, 93891))
        return LinearPointTransform(A)

    tol = 1e-4
    pt = xy_(0, 0)
    for sx, sy in [(3, 4), (0.4, 0.333)]:
        tr = mk_transform(sx, sy)
        sx_, sy_ = get_scale_at_point(pt, tr).xy
        assert abs(sx - sx_) < tol
        assert abs(sy - sy_) < tol

        sx_, sy_ = get_scale_at_point(pt, tr, 0.1).xy
        assert abs(sx - sx_) < tol
        assert abs(sy - sy_) < tol


def test_pix_transform():
    pt = tuple(
        int(x / 10) * 10
        for x in geom.point(145, -35, epsg4326).to_crs(epsg3577).coords[0]
    )

    A = mkA(scale=(20, -20), translation=pt)

    src = GeoBox((512, 1024), A, epsg3577)
    dst = GeoBox.from_geopolygon(src.geographic_extent, resyx_(0.0001, -0.0001))

    tr = native_pix_transform(src, dst)

    pts_src = [xy_(0, 0), xy_(10, 20), xy_(300, 200)]
    pts_dst = tr(pts_src)
    pts_src_ = tr.back(pts_dst)

    np.testing.assert_almost_equal(stack_xy(pts_src), stack_xy(pts_src_))
    assert tr.linear is None
    assert repr(tr).startswith("GbxPointTransform(")

    # check identity transform
    tr = native_pix_transform(src, src)

    pts_src = [xy_(0, 0), xy_(10, 20), xy_(300, 200)]
    pts_dst = tr(pts_src)
    pts_src_ = tr.back(pts_dst)

    np.testing.assert_almost_equal(stack_xy(pts_src), stack_xy(pts_src_))
    np.testing.assert_almost_equal(stack_xy(pts_src), stack_xy(pts_dst))
    assert tr.linear is not None
    assert tr.back.linear is not None
    assert tr.back.back is tr
    assert repr(tr).startswith("LinearPointTransform(")

    # check scale only change
    tr = native_pix_transform(src, scaled_down_geobox(src, 2))
    pts_dst = tr(pts_src)
    pts_src_ = tr.back(pts_dst)

    assert tr.linear is not None
    assert tr.back.linear is not None
    assert tr.back.back is tr

    np.testing.assert_almost_equal(
        stack_xy(pts_dst), [(pt.x / 2, pt.y / 2) for pt in pts_src]
    )

    np.testing.assert_almost_equal(stack_xy(pts_src), stack_xy(pts_src_))


def test_compute_reproject_roi():
    src = AlbersGS.tile_geobox((15, -40))
    dst = GeoBox.from_geopolygon(
        src.extent.to_crs(epsg3857).buffer(10), resolution=src.resolution
    )

    rr = compute_reproject_roi(src, dst)

    assert rr.roi_src == np.s_[0 : src.height, 0 : src.width]
    assert 0 < rr.scale < 1
    assert all(0 < scale < 1 for scale in rr.scale2.xy)
    assert rr.transform.linear is None
    assert rr.transform.back is not None
    assert rr.transform.back.linear is None

    # check pure translation case
    roi_ = np.s_[113:-100, 33:-10]
    rr = compute_reproject_roi(src, src[roi_])
    assert rr.roi_src == roi_normalise(roi_, src.shape)
    assert rr.scale == 1

    rr = compute_reproject_roi(src, src[roi_], padding=0, align=0)
    assert rr.roi_src == roi_normalise(roi_, src.shape)
    assert rr.scale == 1
    assert rr.scale2.xy == (1, 1)

    # check pure translation case
    roi_ = np.s_[113:-100, 33:-10]
    rr = compute_reproject_roi(src, src[roi_], align=256)

    assert rr.roi_src == np.s_[0 : src.height, 0 : src.width]
    assert rr.scale == 1

    roi_ = np.s_[113:-100, 33:-10]
    rr = compute_reproject_roi(src, src[roi_])

    assert rr.scale == 1
    assert roi_shape(rr.roi_src) == roi_shape(rr.roi_dst)
    assert roi_shape(rr.roi_dst) == src[roi_].shape

    # check pasteable zoom_out
    dst = src.zoom_out(2)
    rr = compute_reproject_roi(src, dst)
    assert rr.paste_ok is True
    assert rr.read_shrink == 2
    assert roi_shape(rr.roi_src) == src.shape
    assert roi_shape(rr.roi_dst) == dst.shape

    # check non-pasteable zoom_out
    dst = src[1:, :].zoom_out(2)
    rr = compute_reproject_roi(src, dst)
    assert rr.paste_ok is False
    assert rr.read_shrink == 2


def test_compute_reproject_roi_paste():
    src = GeoBox(
        wh_(1000, 2000),
        mkA(scale=(10, -10), translation=(10 * 123, -10 * 230)),
        epsg3857,
    )

    def _check(src: GeoBox, dst: GeoBox, rr: ReprojectInfo):
        assert rr.read_shrink >= 1

        if roi_is_empty(rr.roi_src):
            assert roi_is_empty(rr.roi_dst)
            return

        if rr.paste_ok:
            if rr.read_shrink == 1:
                assert roi_shape(rr.roi_src) == roi_shape(rr.roi_dst)
                assert src[rr.roi_src].shape == dst[rr.roi_dst].shape
            else:
                # roi source must align to read scale
                # => round-triping roi to overview and back should not change roi
                assert (
                    scaled_up_roi(
                        scaled_down_roi(rr.roi_src, rr.read_shrink), rr.read_shrink
                    )
                    == rr.roi_src
                )

                src_ = src[rr.roi_src].zoom_out(rr.read_shrink)
                assert src_.shape == dst[rr.roi_dst].shape

        if rr.read_shrink == 1:
            assert rr.scale <= 1.1
        else:
            assert rr.scale >= rr.read_shrink

        if src.crs == dst.crs:
            _src = src[rr.roi_src].extent
            _dst = dst[rr.roi_dst].extent
        else:
            _src = src[rr.roi_src].geographic_extent
            _dst = dst[rr.roi_dst].geographic_extent

        assert _src.intersection(_dst).area > 0

    def _yes(src: GeoBox, dst: GeoBox, **kw):
        rr = compute_reproject_roi(src, dst, **kw)
        assert rr.paste_ok is True
        _check(src, dst, rr)

    def _no_(src: GeoBox, dst: GeoBox, **kw):
        rr = compute_reproject_roi(src, dst, **kw)
        assert rr.paste_ok is False
        _check(src, dst, rr)

    t_ = Affine.translation
    s_ = Affine.scale

    # plain pixel aligned translation
    _yes(src, src)
    _yes(src, src[10:, 29:])
    _yes(src[10:, 29:], src)

    # subpixel translation below threshhold
    _no_(src, src * t_(0.3, 0.3))
    _yes(src, src * t_(0.3, 0.3), ttol=0.5)
    _no_(src, src * t_(0.0, 0.1))
    _yes(src, src * t_(0.0, 0.1), ttol=0.15)
    _no_(src, src * t_(-0.1, 0.0))
    _yes(src, src * t_(-0.1, 0.0), ttol=0.15)

    # tiny scale deviations
    _no_(src, src[20:, :30] * s_(1.003, 1.003))
    _yes(src, src[20:, :30] * s_(1.003, 1.003), stol=0.01)

    # integer shrink
    _no_(src, src.zoom_out(2.3))
    _yes(src, src.zoom_out(2))
    _yes(src, src.zoom_out(3))
    _yes(src, src.zoom_out(2 + 1e-5))  # rounding issues should not matter
    _no_(src.zoom_out(3), src)
    _no_(src.zoom_out(2), src)

    # integer shrink but with sub-pixel translation after shrinking
    _yes(src, src[4:, 8:].zoom_out(4))
    _no_(src, src[2:, 8:].zoom_out(4))
    _no_(src, src[8:, 3:].zoom_out(4))
    _yes(src, src[8:, 3:].zoom_out(4), ttol=0.5)


def test_compute_reproject_roi_issue647():
    """In some scenarios non-overlapping geoboxes will result in non-empty
    `roi_dst` even though `roi_src` is empty.

    Test this case separately.
    """

    src = GeoBox(
        (10980, 10980), Affine(10, 0, 300000, 0, -10, 5900020), CRS("epsg:32756")
    )

    dst = GeoBox((976, 976), Affine(10, 0, 1730240, 0, -10, -4170240), CRS("EPSG:3577"))

    assert src.extent.overlaps(dst.extent.to_crs(src.crs)) is False

    rr = compute_reproject_roi(src, dst)

    assert roi_is_empty(rr.roi_src)
    assert roi_is_empty(rr.roi_dst)


def test_compute_reproject_roi_issue1047():
    """`compute_reproject_roi(geobox, geobox[roi])` sometimes returns
    `src_roi != roi`, when `geobox` has (1) tiny pixels and (2) oddly
    sized `alignment`.

    Test this issue is resolved.
    """
    geobox = GeoBox(
        (3000, 3000),
        Affine(
            0.00027778, 0.0, 148.72673054908861, 0.0, -0.00027778, -34.98825802556622
        ),
        "EPSG:4326",
    )
    src_roi = np.s_[2800:2810, 10:30]
    rr = compute_reproject_roi(geobox, geobox[src_roi])

    assert rr.paste_ok is True
    assert rr.roi_src == src_roi
    assert rr.roi_dst == np.s_[0:10, 0:20]


def test_compute_reproject_roi_overhang():
    """
    Images with global coverage in epsg:4326 often have slightly
    wrong georegistration that causes image boundaries to reach outside
    of the [-180, -90, 180, 90] bounding box.

    Reproject roi introduces clipping to deal with that issue.
    """
    tol = 1e-3
    src_geobox = GeoBox.from_bbox(
        (-180 - tol, -90 - tol, 180 + tol, 90 + tol),
        epsg4326,
        shape=wh_(2000, 1000),
        tight=True,
    )
    assert src_geobox.shape.wh == (2000, 1000)
    assert src_geobox.extent.boundingbox[0] < -180
    assert src_geobox.extent.boundingbox[1] < -90
    assert src_geobox.extent.boundingbox[2] > +180
    assert src_geobox.extent.boundingbox[3] > +90

    dst_geobox = GridSpec.web_tiles(0)[0, 0]

    rr = compute_reproject_roi(src_geobox, dst_geobox)
    assert rr.paste_ok is False
    assert dst_geobox[rr.roi_dst] == dst_geobox


def test_axis_overlap():
    s_ = np.s_

    # Source overlaps destination fully
    #
    # S: |<--------------->|
    # D:      |<----->|
    assert compute_axis_overlap(100, 20, 1, 10) == s_[10:30, 0:20]
    assert compute_axis_overlap(100, 20, 2, 10) == s_[10:50, 0:20]
    assert compute_axis_overlap(100, 20, 0.25, 10) == s_[10:15, 0:20]
    assert compute_axis_overlap(100, 20, -1, 80) == s_[60:80, 0:20]
    assert compute_axis_overlap(100, 20, -0.5, 50) == s_[40:50, 0:20]
    assert compute_axis_overlap(100, 20, -2, 90) == s_[50:90, 0:20]

    # Destination overlaps source fully
    #
    # S:      |<-------->|
    # D: |<----------------->|
    assert compute_axis_overlap(10, 100, 1, -10) == s_[0:10, 10:20]
    assert compute_axis_overlap(10, 100, 2, -10) == s_[0:10, 5:10]
    assert compute_axis_overlap(10, 100, 0.5, -10) == s_[0:10, 20:40]
    assert compute_axis_overlap(10, 100, -1, 11) == s_[0:10, 1:11]

    # Partial overlaps
    #
    # S: |<----------->|
    # D:     |<----------->|
    assert compute_axis_overlap(10, 10, 1, 3) == s_[3:10, 0:7]
    assert compute_axis_overlap(10, 15, 1, 3) == s_[3:10, 0:7]

    # S:     |<----------->|
    # D: |<----------->|
    assert compute_axis_overlap(10, 10, 1, -5) == s_[0:5, 5:10]
    assert compute_axis_overlap(50, 10, 1, -5) == s_[0:5, 5:10]

    # No overlaps
    # S: |<--->|
    # D:         |<--->|
    assert compute_axis_overlap(10, 10, 1, 11) == s_[10:10, 0:0]
    assert compute_axis_overlap(10, 40, 1, 11) == s_[10:10, 0:0]

    # S:         |<--->|
    # D: |<--->|
    assert compute_axis_overlap(10, 10, 1, -11) == s_[0:0, 10:10]
    assert compute_axis_overlap(40, 10, 1, -11) == s_[0:0, 10:10]


def test_can_paste():
    assert _can_paste(mkA(translation=(10, -20))) == (True, None)
    assert _can_paste(mkA(scale=(10, 10))) == (True, None)
    assert _can_paste(mkA(scale=(-10, 10), translation=(0, -4 * 10))) == (True, None)

    assert _can_paste(mkA(shear=0.3)) == (False, "has rotation or shear")
    assert _can_paste(mkA(rot=30)) == (False, "has rotation or shear")

    assert _can_paste(mkA(scale=(-11.1, 11.1))) == (False, "non-integer scale")
    assert _can_paste(mkA(scale=(0.5, 0.5))) == (False, "non-integer scale")
    assert _can_paste(mkA(scale=(2, 3))) == (False, "sx!=sy, probably")

    assert _can_paste(mkA(scale=(-10, 10), translation=(0, -4))) == (
        False,
        "sub-pixel translation",
    )
    assert _can_paste(mkA(scale=(-10, 10), translation=(-4, 10))) == (
        False,
        "sub-pixel translation",
    )
    assert _can_paste(mkA(translation=(0, 0.4))) == (False, "sub-pixel translation")
    assert _can_paste(mkA(translation=(0.4, 0))) == (False, "sub-pixel translation")


def test_compute_output_geobox():
    # sentinel2 over Gibraltar strait
    src = GeoBox.from_bbox(
        [199980, 3890220, 309780, 4000020], "EPSG:32630", resolution=10
    )

    # just copy resolution since both in meters
    dst = compute_output_geobox(src, "epsg:6933")
    assert dst.crs.units == src.crs.units
    assert dst.crs == "epsg:6933"
    assert dst.resolution == src.resolution
    assert dst.geographic_extent.contains(src.geographic_extent)
    assert compute_output_geobox(src, "epsg:6933") == src.to_crs("epsg:6933")

    assert compute_output_geobox(
        src, "epsg:6933", resolution="auto"
    ) == compute_output_geobox(src, "epsg:6933", resolution="same")

    # force estimation of new resolution
    dst = compute_output_geobox(src, "epsg:6933", resolution="fit")
    assert dst.crs == "epsg:6933"
    assert dst.resolution != src.resolution
    assert dst.resolution.x == -dst.resolution.y
    assert dst.geographic_extent.contains(src.geographic_extent)

    # force specific resolution
    dst = compute_output_geobox(src, "epsg:6933", resolution=101)
    assert dst.crs == "epsg:6933"
    assert dst.resolution == res_(101)
    assert dst.geographic_extent.contains(src.geographic_extent)

    # check identity case
    assert src is compute_output_geobox(src, src.crs)
    assert src is compute_output_geobox(src, src.crs, resolution="same")
    assert src is compute_output_geobox(src, src.crs, resolution="auto")

    # check conversion to lon/lat
    dst = compute_output_geobox(src, "epsg:4326")
    assert dst.crs == "epsg:4326"
    assert dst.resolution != src.resolution
    assert dst.resolution.x == -dst.resolution.y
    assert dst.geographic_extent.contains(src.geographic_extent)
    npix_change = (src.shape[0] * src.shape[1]) / (dst.shape[0] * dst.shape[1])
    assert 0.8 < npix_change < 1.1

    # check anchor= having effect
    assert compute_output_geobox(
        src, 4326, anchor=AnchorEnum.CENTER
    ) != compute_output_geobox(src, 4326, anchor=AnchorEnum.EDGE)

    # check shape= having effect
    assert compute_output_geobox(src, 4326, shape=(10, 20)).shape == (10, 20)
    assert max(compute_output_geobox(src, 4326, shape=100, tight=True).shape) == 100

    # go back from 4326
    _src = dst
    dst = compute_output_geobox(_src, src.crs)
    npix_change = (_src.shape[0] * _src.shape[1]) / (dst.shape[0] * dst.shape[1])
    assert 0.8 < npix_change < 1.1
    assert dst.geographic_extent.contains(_src.geographic_extent)

    # test bad input
    with pytest.raises(ValueError):
        _ = compute_output_geobox(src, "epsg:6933", resolution="bad-one")



import numpy as np
import numpy.testing as npt
import pytest
import xarray as xr
from affine import Affine

from odc.geo import MaybeCRS
from odc.geo.warp import resampling_s2rio, rio_reproject, rio_warp_affine

NaN = float("nan")


@pytest.mark.parametrize(
    "iso3, crs, resolution",
    [
        ("AUS", "epsg:4326", 0.1),
        ("AUS", "epsg:3577", 10_000),
        ("AUS", "epsg:3857", 10_000),
        ("NZL", "epsg:3857", 5_000),
    ],
)
@pytest.mark.parametrize("resampling", ["nearest", "bilinear", "average", "sum"])
def test_warp_nan(country_raster_f32: xr.DataArray, crs: MaybeCRS, resampling: str):
    xx = country_raster_f32
    assert isinstance(xx, xr.DataArray)
    assert xx.odc.crs == crs
    assert xx.odc.nodata is None
    assert xx.dtype == "float32"

    mid = xx.shape[0] // 2
    xx.data[mid, :] = NaN
    xx.data[:, -10] = NaN

    assert resampling_s2rio(resampling) is not None
    assert np.isnan(xx.data).sum() > 0

    src_gbox = xx.odc.geobox
    dst_gbox = src_gbox.zoom_to(shape=100).pad(10)

    yy1 = np.full(dst_gbox.shape, -333, dtype=xx.dtype)
    yy2 = np.full(dst_gbox.shape, -333, dtype=xx.dtype)

    assert rio_reproject(xx.data, yy1, src_gbox, dst_gbox, resampling=resampling) is yy1
    assert (
        rio_reproject(
            xx.data,
            yy2,
            src_gbox,
            dst_gbox,
            resampling=resampling,
            src_nodata=NaN,
            dst_nodata=NaN,
        )
        is yy2
    )

    npt.assert_array_equal(yy1, yy2)

    # make sure all pixels were replaced
    assert (yy1 == -333).sum() == 0

    # expect to see NaNs in the output
    assert np.isnan(yy2).sum() > 0

    A = Affine.identity()
    xx_ = xx.data.copy() * 0
    assert rio_warp_affine(xx.data, xx_, A, resampling) is xx_
    npt.assert_array_equal(xx.data, xx_)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Write Cloud Optimized GeoTIFFs from xarrays.
"""
from __future__ import annotations

import itertools
from functools import partial
from io import BytesIO
from typing import TYPE_CHECKING, Any, Callable, Optional, Union
from urllib.parse import urlparse
from xml.sax.saxutils import escape as xml_escape

import numpy as np
import xarray as xr


from .._interop import have
from ..geobox import GeoBox
from ..math import resolve_nodata
from ..types import Shape2d, SomeNodata, Unset, shape_
from ._mpu import mpu_write
from ._mpu_fs import MPUFileSink
from ._multipart import MultiPartUploadBase

from ._shared import (
    GDAL_COMP,
    GEOTIFF_TAGS,
    CogMeta,
    compute_cog_spec,
    norm_blocksize,
    yaxis_from_shape,
)

if TYPE_CHECKING:
    import dask.array
    import dask.bag
    from dask.delayed import Delayed

# pylint: disable=too-many-locals,too-many-branches,too-many-arguments,too-many-statements,too-many-instance-attributes


def _render_gdal_metadata(
    band_stats: list[dict[str, float]] | dict[str, float] | None,
    precision: int = 10,
    pad: int = 0,
    eol: str = "",
    gdal_metadata_extra: Optional[list[str]] = None,
) -> str:
    def _item(sample: int, stats: dict[str, float]) -> str:
        return eol.join(
            [
                f'<Item name="STATISTICS_{k.upper()}" sample="{sample:d}">{v:{pad}.{precision}f}</Item>'
                for k, v in stats.items()
            ]
        )

    if band_stats is None:
        band_stats = []
    if isinstance(band_stats, dict):
        band_stats = [band_stats]

    gdal_metadata_extra = [] if gdal_metadata_extra is None else gdal_metadata_extra

    body = eol.join(
        [_item(sample, stats) for sample, stats in enumerate(band_stats)]
        + gdal_metadata_extra
    )
    return eol.join(["<GDALMetadata>", body, "</GDALMetadata>"])


def _unwrap_stats(stats, ndim):
    if ndim == 2:
        return [{k: float(v) for k, v in stats.items()}]

    n = {len(v) for v in stats.values()}.pop()
    return [{k: v[idx] for k, v in stats.items()} for idx in range(n)]


def _stats_from_layer(
    pix: "dask.array.Array", nodata=None, yaxis: int = 0
) -> "Delayed":
    # pylint: disable=import-outside-toplevel
    from dask import array as da
    from dask import delayed

    unwrap = delayed(_unwrap_stats, pure=True, traverse=True)

    axis = (yaxis, yaxis + 1)
    npix = pix.shape[yaxis] * pix.shape[yaxis + 1]

    if nodata is None or np.isnan(nodata):
        dd = pix
        return unwrap(
            {
                "minimum": da.nanmin(dd, axis=axis),
                "maximum": da.nanmax(dd, axis=axis),
                "mean": da.nanmean(dd, axis=axis),
                "stddev": da.nanstd(dd, axis=axis),
                "valid_percent": da.isfinite(dd).sum(axis=axis) * (100 / npix),
            },
            pix.ndim,
        )

    # Exclude both nodata and invalid (e.g. NaN) values from statistics computation
    dd = da.ma.masked_where((pix == nodata) | ~(np.isfinite(pix)), pix)
    return unwrap(
        {
            "minimum": dd.min(axis=axis),
            "maximum": dd.max(axis=axis),
            "mean": dd.mean(axis=axis),
            "stddev": dd.std(axis=axis),
            "valid_percent": da.isfinite(dd).sum(axis=axis) * (100 / npix),
        },
        pix.ndim,
    )


def _make_empty_cog(
    shape: tuple[int, ...],
    dtype: Any,
    gbox: Optional[GeoBox] = None,
    *,
    nodata: SomeNodata = "auto",
    gdal_metadata: Optional[str] = None,
    compression: Union[str, Unset] = Unset(),
    compressionargs: Any = None,
    predictor: Union[int, bool, Unset] = Unset(),
    blocksize: Union[int, list[Union[int, tuple[int, int]]]] = 2048,
    bigtiff: bool = True,
    **kw,
) -> tuple[CogMeta, memoryview]:
    # pylint: disable=import-outside-toplevel,import-error
    have.check_or_error("tifffile", "rasterio", "xarray")
    from tifffile import (
        COMPRESSION,
        FILETYPE,
        PHOTOMETRIC,
        PLANARCONFIG,
        TiffWriter,
        enumarg,
    )

    nodata = resolve_nodata(nodata, dtype)

    predictor, compression, compressionargs = _norm_compression_tifffile(
        dtype,
        predictor,
        compression=compression,
        compressionargs=compressionargs,
        kw=kw,
    )
    _compression = enumarg(COMPRESSION, compression.upper())

    if isinstance(blocksize, int):
        blocksize = [blocksize]

    ax, yaxis = yaxis_from_shape(shape, gbox)
    im_shape = shape_(shape[yaxis : yaxis + 2])
    photometric = PHOTOMETRIC.MINISBLACK
    planarconfig: Optional[PLANARCONFIG] = PLANARCONFIG.SEPARATE
    if ax == "YX":
        nsamples = 1
    elif ax == "YXS":
        nsamples = shape[-1]
        planarconfig = PLANARCONFIG.CONTIG
        if nsamples in (3, 4):
            photometric = PHOTOMETRIC.RGB
    else:
        nsamples = shape[0]
        if nsamples == 1:
            planarconfig = None

    buf = BytesIO()

    opts_common = {
        "dtype": dtype,
        "photometric": photometric,
        "planarconfig": planarconfig,
        "predictor": predictor,
        "compression": _compression,
        "compressionargs": compressionargs,
        "software": False,
        **kw,
    }

    def _sh(shape: Shape2d) -> tuple[int, ...]:
        if ax == "YX":
            return shape.shape
        if ax == "YXS":
            return (*shape.shape, nsamples)
        return (nsamples, *shape.shape)

    tsz = norm_blocksize(blocksize[-1])
    im_shape, _, nlevels = compute_cog_spec(im_shape, tsz)

    extratags: list[tuple[int, int, int, Any]] = []
    if gbox is not None:
        gbox = gbox.expand(im_shape)
        extratags, _ = geotiff_metadata(
            gbox, nodata=nodata, gdal_metadata=gdal_metadata
        )
    # TODO: support nodata/gdal_metadata without gbox?

    _blocks = itertools.chain(iter(blocksize), itertools.repeat(blocksize[-1]))

    tw = TiffWriter(buf, bigtiff=bigtiff, shaped=False)
    metas: list[CogMeta] = []

    for tsz, idx in zip(_blocks, range(nlevels + 1)):
        tile = norm_blocksize(tsz)
        meta = CogMeta(
            ax,
            im_shape,
            shape_(tile),
            nsamples,
            dtype,
            int(_compression),
            predictor,
            compressionargs=compressionargs,
            gbox=gbox,
            nodata=nodata,
        )

        if idx == 0:
            kw = {**opts_common, "extratags": extratags}
        else:
            kw = {**opts_common, "subfiletype": FILETYPE.REDUCEDIMAGE}

        tw.write(
            itertools.repeat(b""),
            shape=_sh(im_shape),
            tile=tile,
            **kw,
        )

        metas.append(meta)
        im_shape = im_shape.shrink2()
        if gbox is not None:
            gbox = gbox.zoom_to(im_shape)

    meta = metas[0]
    meta.overviews = tuple(metas[1:])

    tw.close()

    return meta, buf.getbuffer()


def _cog_block_compressor_yxs(
    block: np.ndarray,
    *,
    tile_shape: tuple[int, ...] = (),
    encoder: Any = None,
    predictor: Any = None,
    fill_value: Union[float, int] = 0,
    **kw,
) -> bytes:
    assert block.ndim == len(tile_shape)
    if tile_shape != block.shape:
        pad = tuple((0, want - have) for want, have in zip(tile_shape, block.shape))
        block = np.pad(block, pad, "constant", constant_values=(fill_value,))

    if predictor is not None:
        block = predictor(block, axis=1)
    if encoder:
        try:
            return encoder(block, **kw)
        except Exception:  # pylint: disable=broad-except
            return b""

    return bytes(block.data)


def _cog_block_compressor_syx(
    block: np.ndarray,
    *,
    tile_shape: tuple[int, int] = (0, 0),
    encoder: Any = None,
    predictor: Any = None,
    fill_value: Union[float, int] = 0,
    sample_idx: int = 0,
    **kw,
) -> bytes:
    assert isinstance(block, np.ndarray)

    if block.ndim == 2:
        pass
    elif block.shape[0] == 1:
        block = block[0, :, :]
    else:
        block = block[sample_idx, :, :]

    assert block.ndim == 2
    if tile_shape != block.shape:
        pad = tuple((0, want - have) for want, have in zip(tile_shape, block.shape))
        block = np.pad(block, pad, "constant", constant_values=(fill_value,))

    if predictor is not None:
        block = predictor(block, axis=1)

    if encoder:
        try:
            return encoder(block, **kw)
        except Exception:  # pylint: disable=broad-except
            return b""

    return bytes(block.data)


def _mk_tile_compressor(
    meta: CogMeta, sample_idx: int = 0
) -> Callable[[np.ndarray], bytes]:
    # pylint: disable=import-outside-toplevel,import-error
    have.check_or_error("tifffile")
    from tifffile import TIFF

    tile_shape = meta.chunks
    encoder = TIFF.COMPRESSORS[meta.compression]

    predictor = None
    if meta.predictor != 1:
        predictor = TIFF.PREDICTORS[meta.predictor]

    fill_value: Union[float, int] = 0
    if meta.nodata is not None:
        fill_value = float(meta.nodata) if isinstance(meta.nodata, str) else meta.nodata

    if meta.axis == "SYX":
        return partial(
            _cog_block_compressor_syx,
            tile_shape=meta.tile.yx,
            encoder=encoder,
            predictor=predictor,
            fill_value=fill_value,
            sample_idx=sample_idx,
            **meta.compressionargs,
        )

    return partial(
        _cog_block_compressor_yxs,
        tile_shape=tile_shape,
        encoder=encoder,
        predictor=predictor,
        fill_value=fill_value,
        **meta.compressionargs,
    )


def _compress_cog_tile(encoder, block, idx):
    return [(encoder(block), idx)]


def _compress_tiles(
    xx: xr.DataArray,
    meta: CogMeta,
    scale_idx: int = 0,
    sample_idx: int = 0,
) -> "dask.bag.Bag":
    """
    Compress chunks according to cog spec.

    :returns: Dask bag of tuples ``(data: bytes, idx: (int, int, int, int))}``
    """
    # pylint: disable=import-outside-toplevel
    have.check_or_error("dask")
    from dask.bag import Bag
    from dask.base import quote, tokenize
    from dask.highlevelgraph import HighLevelGraph

    from .._interop import is_dask_collection

    data = xx.data
    assert is_dask_collection(data)

    if meta.axis == "SYX":
        src_ydim = 1
        if data.ndim == 2:
            _chunks: tuple[int, ...] = meta.tile.yx
        elif len(data.chunks[0]) == 1:
            # if 1 single chunk with all "samples", keep it that way
            _chunks = (data.shape[0], *meta.tile.yx)
        else:
            # else have 1 chunk per "sample"
            _chunks = (1, *meta.tile.yx)

        if data.chunksize != _chunks:
            data = data.rechunk(_chunks)
    else:
        assert meta.num_planes == 1
        src_ydim = 0
        if data.chunksize != meta.chunks:
            data = data.rechunk(meta.chunks)

    encoder = _mk_tile_compressor(meta, sample_idx)

    tk = tokenize(
        data,
        scale_idx,
        meta.axis,
        meta.chunks,
        meta.predictor,
        meta.compression,
        meta.compressionargs,
    )
    cc_id = "" if scale_idx == 0 else f"_{scale_idx}"
    cc_id += "" if meta.num_planes == 1 else f"@{sample_idx}"

    name = f"compress{cc_id}-{tk}"

    src_data_name = data.name

    def block_name(s, y, x):
        if data.ndim == 2:
            return (src_data_name, y, x)
        if src_ydim == 0:
            return (src_data_name, y, x, s)
        if len(data.chunks[0]) == 1:
            return (src_data_name, 0, y, x)
        return (src_data_name, s, y, x)

    dsk: Any = {}
    for i, (s, y, x) in enumerate(meta.tidx(sample_idx)):
        block = block_name(s, y, x)
        dsk[name, i] = (_compress_cog_tile, encoder, block, quote((scale_idx, s, y, x)))

    nparts = len(dsk)
    dsk = HighLevelGraph.from_collections(name, dsk, dependencies=[data])
    return Bag(dsk, name, nparts)


def _pyramids_from_cog_metadata(
    xx: xr.DataArray,
    cog_meta: CogMeta,
    resampling: Union[str, int] = "nearest",
) -> tuple[xr.DataArray, ...]:
    out = [xx]

    for mm in cog_meta.overviews:
        gbox = mm.gbox
        out.append(
            out[-1].odc.reproject(gbox, chunks=mm.tile.yx, resampling=resampling)
        )

    return tuple(out)


def _extract_tile_info(
    meta: CogMeta,
    tiles: list[tuple[int, int, int, int, int]],
    start_offset: int = 0,
) -> list[tuple[list[int], list[int]]]:
    mm = meta.flatten()
    tile_info = [([0] * m.num_tiles, [0] * m.num_tiles) for m in mm]

    byte_offset = start_offset
    for scale_idx, p, y, x, sz in tiles:
        m = mm[scale_idx]
        b_offsets, b_lengths = tile_info[scale_idx]

        tidx = m.flat_tile_idx((p, y, x))
        if sz != 0:
            b_lengths[tidx] = sz
            b_offsets[tidx] = byte_offset
            byte_offset += sz

    return tile_info


def _patch_hdr(
    tiles: list[tuple[int, tuple[int, int, int, int]]],
    meta: CogMeta,
    hdr0: bytes,
    stats: Optional[list[dict[str, float]]] = None,
    gdal_metadata_extra: Optional[list[str]] = None,
) -> bytes:
    # pylint: disable=import-outside-toplevel,import-error
    from tifffile import TiffFile, TiffPage

    _tiles = [(*idx, sz) for sz, idx in tiles]
    tile_info = _extract_tile_info(meta, _tiles, 0)

    _bio = BytesIO(hdr0)
    with TiffFile(_bio, mode="r+", name=":mem:") as tr:
        assert len(tile_info) == len(tr.pages)
        if stats is not None or gdal_metadata_extra:
            md_tag = tr.pages.first.tags.get(42112, None)
            assert md_tag is not None
            gdal_metadata = _render_gdal_metadata(
                stats, precision=6, gdal_metadata_extra=gdal_metadata_extra
            )
            md_tag.overwrite(gdal_metadata)

        hdr_sz = len(_bio.getbuffer())

        # 324 -- offsets
        # 325 -- byte counts
        for info, page in zip(tile_info, tr.pages):
            assert isinstance(page, TiffPage)
            tags = page.tags
            offsets, lengths = info
            tags[324].overwrite([off + hdr_sz for off in offsets])
            tags[325].overwrite(lengths)

    return bytes(_bio.getbuffer())


def _norm_predictor(predictor: Union[int, bool, None], dtype: Any) -> int:
    if predictor is False or predictor is None:
        return 1

    if predictor is True:
        dtype = np.dtype(dtype)
        if dtype.kind == "f":
            return 3
        if dtype.kind in "ui" and dtype.itemsize <= 4:
            return 2
        return 1
    return predictor


def _norm_compression_tifffile(
    dtype: Any,
    predictor: Union[bool, None, int, Unset] = Unset(),
    compression: Union[str, Unset] = Unset(),
    compressionargs: Any = None,
    level: Optional[Union[int, float]] = None,
    kw: Optional[dict[str, Any]] = None,
) -> tuple[int, str, dict[str, Any]]:
    if kw is None:
        kw = {}
    if isinstance(compression, Unset):
        compression = kw.pop("compress", "ADOBE_DEFLATE")
        assert isinstance(compression, str)

    if compressionargs is None:
        compressionargs = {}

    remap = {k.upper(): k for k in kw}

    def opt(name: str, default=None) -> Any:
        k = remap.get(name.upper(), None)
        if k is None:
            return default
        return kw.pop(k, default)

    def _gdal_level(compression: str, default=None) -> Any:
        gdal_level_k = GDAL_COMP.get(compression, None)
        if gdal_level_k is None:
            return default
        return opt(gdal_level_k, default)

    compression = compression.upper()

    if level is None and "level" not in compressionargs:
        # GDAL compat
        level = _gdal_level(compression)

    if level is not None:
        compressionargs["level"] = level

    if compression == "DEFLATE":
        compression = "ADOBE_DEFLATE"
    if compression == "LERC_DEFLATE":
        compression = "LERC"
        compressionargs["compression"] = "deflate"
        if (lvl := _gdal_level("DEFLATE")) is not None:
            compressionargs["compressionargs"] = {"level": lvl}
    elif compression == "LERC_ZSTD":
        compression = "LERC"
        compressionargs["compression"] = "zstd"
        if (lvl := _gdal_level("ZSTD")) is not None:
            compressionargs["compressionargs"] = {"level": lvl}

    if isinstance(predictor, Unset):
        predictor = compression in ("ADOBE_DEFLATE", "ZSTD", "LZMA")

    predictor = _norm_predictor(predictor, dtype)
    return (predictor, compression, compressionargs)


def _gdal_sample_description(sample: int, description: str) -> str:
    """Make XML line of GDAL metadata.

    :param band: Sample / band number in data array.
    :param description: Band name in data array.

    :return: GDAL XML metadata line to place in TIFF file.
    """
    # GDAL does double escaping; see frmts/gtiff/geotiff.cpp.
    # We also double escape to maximize compatibility with tools expecting GDAL-generated metadata.
    double_escaped_description = xml_escape(xml_escape(description))
    return f'<Item name="DESCRIPTION" sample="{sample}" role="description">{double_escaped_description}</Item>'


def _band_names(xx: xr.DataArray) -> list[str]:
    if "band" in xx.coords and xx.coords["band"].dtype.type is np.str_:
        return list(xx["band"].values)
    if "long_name" in xx.attrs:
        long_name = xx.attrs["long_name"]
        return [long_name] if isinstance(long_name, str) else long_name
    return []


def _gdal_sample_descriptions(descriptions: list[str]) -> list[str]:
    """Convert band names to GDAL sample descriptions.

    :return: List of GDAL XML metadata lines to place in TIFF file.
    """
    return [
        _gdal_sample_description(sample, description)
        for sample, description in enumerate(descriptions)
    ]


def save_cog_with_dask(
    xx: xr.DataArray,
    dst: str = "",
    *,
    compression: Union[str, Unset] = Unset(),
    compressionargs: Any = None,
    level: Optional[Union[int, float]] = None,
    predictor: Union[int, bool, Unset] = Unset(),
    blocksize: Union[Unset, int, list[Union[int, tuple[int, int]]]] = Unset(),
    bigtiff: bool = True,
    overview_resampling: Union[int, str] = "nearest",
    aws: Optional[dict[str, Any]] = None,
    azure: Optional[dict[str, Any]] = None,
    client: Any = None,
    stats: bool | int = True,
    **kw,
) -> Any:
    """
    Save a Cloud Optimized GeoTIFF to S3, Azure Blob Storage, or file with Dask.

    :param xx: Pixels as :py:class:`xarray.DataArray` backed by Dask
    :param dst: S3, Azure URL, or file path
    :param compression: Compression to use, default is ``DEFLATE``
    :param level: Compression "level", depends on chosen compression
    :param predictor: TIFF predictor setting
    :param compressionargs: Any other compression arguments
    :param overview_resampling: Resampling to use for computing overviews
    :param blocksize: Configure blocksizes for main and overview images
    :param bigtiff: Generate BigTIFF by default, set to ``False`` to disable
    :param aws: Configure AWS write access
    :param azure: Azure credentials/config
    :param client: Dask client
    :param stats: Set to ``False`` to disable stats computation

    :returns: Dask delayed
    """
    # pylint: disable=import-outside-toplevel
    import dask.bag

    from ..xr import ODCExtensionDa

    aws = aws or {}
    azure = azure or {}

    upload_params = {
        k: kw.pop(k, None) for k in ["writes_per_chunk", "spill_sz"] if k in kw
    }
    parts_base = kw.pop("parts_base", None)

    # Normalise compression settings and remove GDAL compat options from kw
    predictor, compression, compressionargs = _norm_compression_tifffile(
        xx.dtype, predictor, compression, compressionargs, level=level, kw=kw
    )

    xx_odc = xx.odc
    assert isinstance(xx_odc, ODCExtensionDa)
    assert isinstance(xx_odc.geobox, GeoBox) or xx_odc.geobox is None

    ydim = xx_odc.ydim
    data_chunks: tuple[int, int] = xx.data.chunksize[ydim : ydim + 2]
    if isinstance(blocksize, Unset):
        blocksize = [data_chunks, int(max(*data_chunks) // 2)]

    # Metadata
    band_names = _band_names(xx)
    sample_descriptions_metadata = _gdal_sample_descriptions(band_names)
    gdal_metadata = None if stats is False and not band_names else ""

    # Prepare COG metadata and header
    meta, hdr0 = _make_empty_cog(
        xx.shape,
        xx.dtype,
        xx_odc.geobox,
        predictor=predictor,
        compression=compression,
        compressionargs=compressionargs,
        blocksize=blocksize,
        bigtiff=bigtiff,
        nodata=xx_odc.nodata,
        gdal_metadata=gdal_metadata,
        **kw,
    )
    hdr0 = bytes(hdr0)

    if band_names and len(band_names) != meta.nsamples:
        raise ValueError(
            f"Found {len(band_names)} band names ({band_names}), expected {meta.nsamples} bands."
        )

    layers = _pyramids_from_cog_metadata(xx, meta, resampling=overview_resampling)

    if stats is True:
        stats = len(layers) // 2

    _stats: "Delayed" | None = None
    if stats is not False:
        _stats = _stats_from_layer(
            layers[stats].data, nodata=xx_odc.nodata, yaxis=xx_odc.ydim
        )

    # Prepare tiles
    _tiles: list["dask.bag.Bag"] = []
    for scale_idx, (mm, img) in enumerate(zip(meta.flatten(), layers)):
        for sample_idx in range(meta.num_planes):
            tt = _compress_tiles(img, mm, scale_idx=scale_idx, sample_idx=sample_idx)
            if tt.npartitions > 20:
                tt = tt.repartition(npartitions=tt.npartitions // 4)
            _tiles.append(tt)

    if dst == "":
        return {
            "meta": meta,
            "hdr0": hdr0,
            "tiles": _tiles,
            "layers": layers,
            "_stats": _stats,
        }

    # Determine output type and initiate uploader
    parsed_url = urlparse(dst)
    if parsed_url.scheme == "s3":
        if have.botocore:
            from ._s3 import S3MultiPartUpload, s3_parse_url

            bucket, key = s3_parse_url(dst)
            uploader: MultiPartUploadBase = S3MultiPartUpload(bucket, key, **aws)
        else:
            raise RuntimeError("Please install `boto3` to use S3")
    elif parsed_url.scheme == "az":
        if have.azure:
            from ._az import AzMultiPartUpload

            assert azure is not None
            assert "account_url" in azure
            assert "credential" in azure

            uploader = AzMultiPartUpload(
                account_url=azure["account_url"],
                container=parsed_url.netloc,
                blob=parsed_url.path.lstrip("/"),
                credential=azure["credential"],
            )
        else:
            raise RuntimeError("Please install `azure-storage-blob` to use Azure")
    else:
        # Assume local disk
        write = MPUFileSink(dst, parts_base=parts_base)
        return mpu_write(
            _tiles[::-1],
            write,
            mk_header=_patch_hdr,
            user_kw={
                "meta": meta,
                "hdr0": hdr0,
                "stats": _stats,
                "gdal_metadata_extra": sample_descriptions_metadata,
            },
            **upload_params,
        )

    # Upload tiles
    tiles_write_order = _tiles[::-1]  # Reverse tiles for writing
    if len(tiles_write_order) > 4:  # Optimize for larger datasets
        tiles_write_order = [
            dask.bag.concat(tiles_write_order[:4]),
            *tiles_write_order[4:],
        ]

    return uploader.upload(
        tiles_write_order,
        mk_header=_patch_hdr,
        user_kw={
            "meta": meta,
            "hdr0": hdr0,
            "stats": _stats,
            "gdal_metadata_extra": sample_descriptions_metadata,
        },
        client=client,
        **upload_params,
    )


def geotiff_metadata(
    geobox: GeoBox,
    nodata: SomeNodata = "auto",
    gdal_metadata: Optional[str] = None,
) -> tuple[list[tuple[int, int, int, Any]], dict[str, Any]]:
    """
    Convert GeoBox to geotiff tags and metadata for :py:mod:`tifffile`.

    .. note::

       Requires :py:mod:`rasterio`, :py:mod:`tifffile` and :py:mod:`xarray`.


    :returns:
       List of TIFF tag tuples suitable for passing to :py:mod:`tifffile` as
       ``extratags=``, and dictionary representation of GEOTIFF tags.

    """
    # pylint: disable=import-outside-toplevel,import-error

    if not (have.tifffile and have.rasterio):
        raise RuntimeError(
            "Please install `tifffile` and `rasterio` to use this method"
        )

    from tifffile import TiffFile

    from ..xr import xr_zeros
    from ._rio import to_cog

    buf = to_cog(
        xr_zeros(geobox[:2, :2]), nodata=nodata, compress=None, overview_levels=[]
    )
    tf = TiffFile(BytesIO(buf), mode="r")
    assert tf.geotiff_metadata is not None

    def _dtype_as_int(dtype) -> int:
        if isinstance(dtype, int):
            return dtype
        return dtype.value

    geo_tags: list[tuple[int, int, int, Any]] = [
        (t.code, _dtype_as_int(t.dtype), t.count, t.value)
        for t in tf.pages.first.tags.values()
        if t.code in GEOTIFF_TAGS
    ]

    if gdal_metadata is not None:
        geo_tags.append((42112, 2, len(gdal_metadata) + 1, gdal_metadata))

    return geo_tags, tf.geotiff_metadata



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2023 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Working with 2d+ chunks.
"""
from typing import Any, Iterator, Mapping, Optional, Tuple

import numpy as np

from .roi import VariableSizedTiles, roi_intersect3, roi_normalise, roi_shape
from .types import Chunks2d


def _find_common_type(array_types, scalar_type=None):
    if scalar_type is None:
        return np.result_type(*array_types)

    if np.issubdtype(scalar_type, np.floating):
        array_types = array_types + [0.0]
    elif np.issubdtype(scalar_type, np.complexfloating):
        array_types = array_types + [0j]
    return np.result_type(*array_types)


class BlockAssembler:
    """
    Construct contigous 2+ dim array from incomplete set of YX blocks.
    """

    def __init__(
        self,
        blocks: Mapping[Tuple[int, int], np.ndarray],
        chunks: Chunks2d,
        axis: int = 0,
    ) -> None:
        self._shape = BlockAssembler._verify_shape(blocks, chunks, axis=axis)
        self._dtype = (
            np.dtype("float32")
            if len(blocks) == 0
            else _find_common_type([b.dtype for b in blocks.values()])
        )
        self._axis = axis
        self._blocks = blocks
        self._tiles = VariableSizedTiles(chunks)
        assert self._tiles.base.shape == self._shape[axis : axis + 2]

    @staticmethod
    def _verify_shape(
        blocks: Mapping[Tuple[int, int], np.ndarray],
        chunks: Chunks2d,
        axis: int = 0,
    ) -> Tuple[int, ...]:
        """
        compute total shape including prefix and postfix dimensions.
        """
        state: Optional[Tuple[int, Tuple[int, ...], Tuple[int, ...]]] = None
        chy, chx = chunks
        for (iy, ix), b in blocks.items():
            if state is None:
                if b.ndim < axis + 2:
                    raise ValueError(
                        f"Too few dimensions for `axis={axis}` ({b.ndim} < {axis+2})"
                    )
                state = (b.ndim, b.shape[:axis], b.shape[axis + 2 :])

            if state[0] != b.ndim or state[1:] != (b.shape[:axis], b.shape[axis + 2 :]):
                raise ValueError("Extra dimensions must be the same across all blocks")

            yx_shape = b.shape[axis : axis + 2]
            if yx_shape != (chy[iy], chx[ix]):
                raise ValueError(f"Mismatched block YxX shape at [{iy}, {ix}]")

        ny, nx = sum(chy), sum(chx)
        if state is None:
            return (ny, nx)
        _, prefix, postfix = state
        return (*prefix, ny, nx, *postfix)

    @property
    def shape(self) -> Tuple[int, ...]:
        return self._shape

    @property
    def dtype(self) -> np.dtype:
        return self._dtype

    @property
    def ndim(self) -> int:
        return len(self._shape)

    def _norm_roi(self, roi):
        if roi is None:
            roi = tuple(slice(0, n) for n in self._shape)
        if not isinstance(roi, tuple):
            roi = (roi,)
        ndim = len(roi)
        if ndim == 2:
            prefix = tuple(slice(0, n) for n in self._shape[: self._axis])
            postfix = tuple(slice(0, n) for n in self._shape[self._axis + 2 :])
            roi = (*prefix, *roi, *postfix)
        elif ndim < self.ndim:
            roi = (*roi, *tuple(slice(0, n) for n in self._shape[ndim:]))
        elif ndim > self.ndim:
            raise IndexError("Too many index dimensions")

        # squeeze out non Y/X axis that were indexed with a single int
        YX = (self._axis, self._axis + 1)
        to_squeze = tuple(
            idx
            for idx, s in enumerate(roi)
            if not isinstance(s, slice) and idx not in YX
        )
        return roi_normalise(roi, self._shape), to_squeze

    def with_yx(self, src, yx):
        return (*src[: self._axis], *yx, *src[self._axis + 2 :])

    def extract(
        self,
        fill_value: Any = None,
        *,
        dtype=None,
        roi=None,
        casting="same_kind",
    ) -> np.ndarray:
        """
        Paste all blocks together into one array possibly with type coercion.
        """
        if dtype is None:
            dtype = self._dtype
            if fill_value is not None:
                # possibly upgrade to float based on fill_value
                dtype = _find_common_type([dtype], np.min_scalar_type(fill_value))
        else:
            dtype = np.dtype(dtype)

        if fill_value is None:
            fill_value = dtype.type("nan" if np.issubdtype(dtype, np.floating) else 0)

        roi, squeeze_axis = self._norm_roi(roi)
        assert len(roi) == self.ndim

        yx_roi = roi[self._axis : self._axis + 2]
        everything = tuple(slice(None) for _ in range(self.ndim))

        xx = np.full(roi_shape(roi), fill_value, dtype=dtype)

        for idx, block in self._blocks.items():
            yx_roi_b = self._tiles[idx]  # area covered by this block
            s_roi, d_roi, _ = roi_intersect3(yx_roi_b, yx_roi)
            s_roi = self.with_yx(roi, s_roi)
            d_roi = self.with_yx(everything, d_roi)
            np.copyto(xx[d_roi], block[s_roi], casting=casting)

        if squeeze_axis:
            xx = np.squeeze(xx, axis=squeeze_axis)
        return xx

    def __getitem__(self, roi) -> np.ndarray:
        return self.extract(roi=roi)

    def planes_yx(self, yx_roi=None) -> Iterator[Any]:
        a = self._axis
        if yx_roi is None:
            yx = np.s_[:, :]
        else:
            ry, rx = yx_roi
            yx = (ry, rx)

        for idx in np.ndindex(self.shape[:a] + self.shape[a + 2 :]):
            yield (*idx[:a], *yx, *idx[a:])



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Xarray Intergration.

Importing this module enables `.odc.` extension on :py:class:`xarray.DataArray` and
:py:class:`xarray.Dataset` objects.

For more details see: :py:class:`~odc.geo.xr.ODCExtension` and
:py:class:`~odc.geo.xr.ODCExtensionDs`.
"""
from ._interop import have
from ._xr_interop import (
    ODCExtension,
    ODCExtensionDa,
    ODCExtensionDs,
    assign_crs,
    colorize,
    crop,
    mask,
    rasterize,
    register_geobox,
    spatial_dims,
    to_rgba,
    wrap_xr,
    xr_coords,
    xr_reproject,
    xr_zeros,
)

wrap = wrap_xr

__all__ = [
    "ODCExtension",
    "ODCExtensionDa",
    "ODCExtensionDs",
    "assign_crs",
    "rasterize",
    "register_geobox",
    "spatial_dims",
    "xr_coords",
    "wrap",
    "wrap_xr",
    "xr_reproject",
    "xr_zeros",
    "colorize",
    "to_rgba",
    "crop",
    "mask",
]

# pylint: disable=import-outside-toplevel,unused-import
if have.rasterio:
    from ._xr_interop import add_to, compress, rio_reproject, to_cog, write_cog, explore

    __all__.extend(
        [
            "to_cog",
            "write_cog",
            "add_to",
            "rio_reproject",
            "compress",
            "explore",
        ]
    )



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Various mathy helpers.

Minimal dependencies in this module.
"""
from __future__ import annotations

from math import ceil, floor, fmod, isfinite, log2
from typing import (
    Any,
    Iterator,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

import numpy as np
from affine import Affine
from numpy.polynomial.polynomial import polygrid2d, polyval2d

from ._interop import is_dask_collection
from .types import (
    XY,
    AnchorEnum,
    FillValue,
    Nodata,
    Resolution,
    SomeNodata,
    SomeResolution,
    SomeShape,
    res_,
    resxy_,
    shape_,
    xy_,
)

AffineX = TypeVar("AffineX", np.ndarray, Affine)


def maybe_zero(x: float, tol: float) -> float:
    """Turn almost zeros to actual zeros."""
    if abs(x) < tol:
        return 0
    return x


def split_float(x: float) -> Tuple[float, float]:
    """
    Split float number into whole and fractional parts.

    Adding the two numbers back together should result in the original value.
    Fractional part is always in the ``(-0.5, +0.5)`` interval, and whole part
    is equivalent to ``round(x)``.

    :param x: floating point number
    :return: ``whole, fraction``
    """
    if not isfinite(x):
        return (x, 0)

    x_part = fmod(x, 1.0)
    x_whole = x - x_part
    if x_part > 0.5:
        x_part -= 1
        x_whole += 1
    elif x_part < -0.5:
        x_part += 1
        x_whole -= 1
    return (x_whole, x_part)


def maybe_int(x: float, tol: float) -> Union[int, float]:
    """
    Turn almost ints to actual ints.

    pass through other values unmodified.
    """
    if not isfinite(x):
        return x

    x_whole, x_part = split_float(x)

    if abs(x_part) < tol:  # almost int
        return int(x_whole)
    return x


def snap_scale(s: float, tol: float = 1e-6) -> float:
    """
    Snap scale.

    Snap scale to the nearest integer or simple fractions in the form ``1/<int>`` if within
    tolerance.

    :return: ``s`` if too far from snap
    :return: snapped version of ``s`` when within ``tol`` of the integer scale.
    """
    if abs(s) >= 1 - tol:
        return maybe_int(s, tol)

    # Check of s is 0
    if abs(s) < tol:
        return s

    # Check for simple fractions
    s_inv = 1 / s
    s_inv_snapped = maybe_int(s_inv, tol)
    if s_inv_snapped is s_inv:
        return s
    return 1 / s_inv_snapped


def align_down(x: int, align: int) -> int:
    """
    Align integer down.

    :return:
      ``y`` such that ``y % align == 0`` and ``y <= x`` and ``(x - y) < align``
    """
    return x - (x % align)


def align_up(x: int, align: int) -> int:
    """
    Align integer up.

    :return:
      ``y`` such that ``y % align == 0`` and ``y >= x`` and ``(y - x) < align``
    """
    return align_down(x + (align - 1), align)


def align_up_pow2(x: int) -> int:
    """
    Align up to the nearest power of 2.

    :return:
       Smallest ``y`` such that ``y >= x`` and ``y == 2**n`` for some integer ``n``.
    """
    if x <= 0:
        return 1
    return 2 ** int(ceil(log2(x)))


def align_down_pow2(x: int) -> int:
    """
    Align down to the nearest power of 2.

    :return:
       Largest ``y`` such that ``y <= x`` and ``y == 2**n`` for some integer ``n``.
    """
    y = align_up_pow2(x)
    if y > x:
        y = y // 2
    return y


def clamp(x, lo, up):
    """Clamp ``x`` to be ``lo <= x <= up``."""
    assert lo <= up
    return lo if x < lo else up if x > up else x


def is_almost_int(x: float, tol: float) -> bool:
    """
    Check if number is close enough to an integer.

    :param x: number to check
    :param tol: tolerance to use.
    """
    if not isfinite(x):
        return False

    x = abs(fmod(x, 1))
    if x > 0.5:
        x = 1 - x
    return x < tol


def resolve_fill_value(dst_nodata: Nodata, src_nodata: Nodata, dtype) -> FillValue:
    dtype = np.dtype(dtype)

    if dst_nodata is not None:
        return dtype.type(dst_nodata)
    if np.issubdtype(dtype, np.floating):
        return dtype.type("nan")
    if src_nodata is not None:
        return dtype.type(src_nodata)
    return dtype.type(0)


def resolve_nodata(
    nodata: SomeNodata,
    dtype=None,
    xr_nodata=None,
) -> Nodata:
    # pylint: disable=too-many-return-statements
    if nodata is None:
        return None

    if nodata == "auto":
        if xr_nodata is not None:
            return xr_nodata
        if dtype is None:
            return None
        if np.issubdtype(dtype, np.floating):
            return np.nan
        return None

    if isinstance(nodata, str):
        return float(nodata)
    return nodata


def is_nodata_empty(nodata: Nodata) -> bool:
    if nodata is None:
        return True
    if isinstance(nodata, float) and np.isnan(nodata):
        return True
    return False


def _snap_edge_pos(x0: float, x1: float, res: float, tol: float) -> Tuple[float, int]:
    assert res > 0
    assert x1 >= x0
    _x0 = floor(maybe_int(x0 / res, tol))
    _x1 = ceil(maybe_int(x1 / res, tol))
    nx = max(1, _x1 - _x0)
    return _x0 * res, nx


def _snap_edge(x0: float, x1: float, res: float, tol: float) -> Tuple[float, int]:
    assert x1 >= x0
    if res > 0:
        return _snap_edge_pos(x0, x1, res, tol)
    _tx, nx = _snap_edge_pos(x0, x1, -res, tol)
    tx = _tx + nx * (-res)
    return tx, nx


def snap_grid(
    x0: float, x1: float, res: float, off_pix: Optional[float] = 0, tol: float = 1e-6
) -> Tuple[float, int]:
    """
    Compute grid snapping for single axis.

    :param x0: In point ``x0 <= x1``
    :param x1: Out point ``x0 <= x1``
    :param res: Pixel size and direction (can be negative)
    :param off_pix:
       Pixel fraction to align to ``x=0``.
       0 - edge aligned
       0.5 - center aligned
       None - don't snap

    :return: ``tx, nx`` that defines 1-d grid, such that ``x0`` and ``x1`` are within edge pixels.
    """
    assert (off_pix is None) or (0 <= off_pix < 1)
    if off_pix is None:
        if res > 0:
            nx = ceil(maybe_int((x1 - x0) / res, tol))
            return x0, max(1, nx)
        nx = ceil(maybe_int((x1 - x0) / (-res), tol))
        return x1, max(nx, 1)

    off = off_pix * abs(res)
    _tx, nx = _snap_edge(x0 - off, x1 - off, res, tol)
    return _tx + off, nx


def data_resolution_and_offset(
    data, fallback_resolution: Optional[float] = None
) -> Tuple[float, float]:
    """
    Compute resolution and offset from x/y axis data.

    Only uses first two coordinate values, assumes that data is regularly
    sampled.

    :returns: ``(resolution, offset)``
    """
    if is_dask_collection(data):
        data = data.values

    if data.size < 2:
        if data.size < 1:
            raise ValueError("Can't calculate resolution for empty data")
        if fallback_resolution is None:
            raise ValueError("Can't calculate resolution with data size < 2")
        res = fallback_resolution
        off = data[0] - 0.5 * res
        return res, float(off)

    x0, x1 = map(float, data[:2])

    # x0 = 0*s + t + s/2
    # x1 = 1*s + t + s/2
    ####################
    # s   = x1 - x0
    # 2*t = 3*x0 - x1
    return x1 - x0, (3 * x0 - x1) / 2


def affine_from_axis(
    xx, yy, fallback_resolution: Optional[SomeResolution] = None
) -> Affine:
    """
    Compute Affine transform from axis.

    Transform direction is from pixel coordinates to CRS units of ``X/Y`` axis.

    :param xx:
       ``X`` axis coordinates

    :param yy:
       ``Y`` axis coordinates

    :param fallback_resolution:
       Resolution to assume for single element axis.

    .. code-block:: text

        (0,0) in pixel space is defined as top left corner of the top left pixel
          |
          V 0   1
          +---+---+
        0 |   |   |
          +---+---+
        1 |   |   |
          +---+---+

    Only uses first two coordinate values, assumes that data is regularly
    sampled.

    :raises:
       :py:class:`ValueError` when any axis is empty.

    :raises:
       :py:class:`ValueError` when any axis has single value and fallback
       resolution was not supplied.
    """
    frx: Optional[float] = None
    fry: Optional[float] = None

    if fallback_resolution is not None:
        frx, fry = res_(fallback_resolution).xy

    xres, xoff = data_resolution_and_offset(xx, frx)
    yres, yoff = data_resolution_and_offset(yy, fry)

    return Affine(xres, 0.0, xoff, 0.0, yres, yoff)


def apply_affine(
    A: Affine, x: np.ndarray, y: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Broadcast ``A*(x_i, y_i)`` across all elements of ``x/y``.

    Arrays could be in any shape (usually 2d image).

    :param A: Affine transform to apply
    :param x: x coordinates (any shape)
    :param y: y coordinates (same shape as ``x``)
    :returns: Transformed coordinate as two arrays of the same shape as input ``(x', y')``
    """

    shape = x.shape

    A = np.asarray(A).reshape(3, 3)
    t = A[:2, -1].reshape((2, 1))
    A = A[:2, :2]

    x, y = A @ np.vstack([x.ravel(), y.ravel()]) + t
    x, y = (a.reshape(shape) for a in (x, y))
    return (x, y)


def split_translation(t: XY[float]) -> Tuple[XY[float], XY[float]]:
    """
    Split translation into pixel aligned and sub-pixel parts.

    Subpixel translation is guaranteed to be in ``[-0.5, +0.5]`` range.

    .. code-block:: python

       x + t = x + t_whole + t_subpix

    :param t: Translation as :py:class:`~odc.geo.XY`

    :returns: ``(t_whole, t_subpix)``
    """
    _tt = t.map(split_float)
    whole = _tt.map(lambda x: x[0])
    part = _tt.map(lambda x: x[1])
    return whole, part


def is_affine_st(A: Affine, tol: float = 1e-10) -> bool:
    """
    Check if transfrom is pure scale and translation.

    :return: ``True`` if Affine transform has scale and translation components only
    :return: ``False`` if there is non-zero rotation or skew
    """
    (_, wx, _, wy, _, _, *_) = A

    return abs(wx) < tol and abs(wy) < tol


def approx_equal_affine(a: Affine, b: Affine, tol: float = 1e-6) -> bool:
    """
    Check if two affine matrices are approximately equal.
    """
    sx, z1, tx, z2, sy, ty = map(lambda v: maybe_int(v, tol), (~a * b)[:6])
    return (sx, z1, tx, z2, sy, ty) == (1, 0, 0, 0, 1, 0)


def snap_affine(
    A: Affine, ttol: float = 1e-3, stol: float = 1e-6, tol: float = 1e-8
) -> Affine:
    """
    Snap scale and translation parts to integer when close enough.

    When scale is less than 1 then attempt snapping to ``1/<int>``.

    If input has rotation/shear component then return unchanged.

    :param A: Affine matrix
    :param ttol: translation tolerance, defaults to 1e-3
    :param stol: scale tolerance, defaults to 1e-6
    :param tol: rotation tolerance, defaults to 1e-10

    :return: Adjusted Affine matrix.
    :return: Input Affine matrix when rotation/shear is present.
    """
    sx, wx, tx, wy, sy, ty, *_ = A

    # has rotation component
    if abs(wx) > tol or abs(wy) > tol:
        return A

    sx_ = snap_scale(sx, stol)
    sy_ = snap_scale(sy, stol)
    tx_ = maybe_int(tx, ttol)
    ty_ = maybe_int(ty, ttol)
    return Affine(sx_, 0, tx_, 0, sy_, ty_)


def decompose_rws(A: AffineX) -> Tuple[AffineX, AffineX, AffineX]:
    """
     Compute decomposition Affine matrix sans translation into Rotation, Shear and Scale.

     Find matrices ``R,W,S`` such that ``A = R W S`` and

     .. code-block:

        R [ca -sa]  W [1, w]  S [sx,  0]
          [sa  ca]    [0, 1]    [ 0, sy]

    .. note:

       There are ambiguities for negative scales.

       * ``R(90)*S(1,1) == R(-90)*S(-1,-1)``

       * ``(R*(-I))*((-I)*S) == R*S``


     :return: Rotation, Shear, Scale ``2x2`` matrices
    """
    # pylint: disable=too-many-locals

    if isinstance(A, Affine):

        def to_affine(m, t=(0, 0)):
            a, b, d, e = m.ravel()
            c, f = t
            return Affine(a, b, c, d, e, f)

        (a, b, c, d, e, f, *_) = A
        R, W, S = decompose_rws(np.asarray([[a, b], [d, e]], dtype="float64"))

        return to_affine(R, (c, f)), to_affine(W), to_affine(S)

    assert A.shape == (2, 2)

    WS = np.linalg.cholesky(A.T @ A).T
    R = A @ np.linalg.inv(WS)

    if np.linalg.det(R) < 0:
        R[:, -1] *= -1
        WS[-1, :] *= -1

    ss = np.diag(WS)
    S = np.diag(ss)
    W = WS @ np.diag(1.0 / ss)

    return R, W, S


def stack_xy(pts: Sequence[XY[float]]) -> np.ndarray:
    """Turn into an ``Nx2`` ndarray of floats in X,Y order."""
    return np.vstack([pt.xy for pt in pts])


def unstack_xy(pts: np.ndarray) -> List[XY[float]]:
    """Turn ``Nx2`` array in X,Y order into a list of XY points."""
    assert pts.ndim == 2 and pts.shape[1] == 2
    return [xy_(pt) for pt in pts]


def norm_xy(
    pts: np.ndarray, out: Optional[np.ndarray] = None
) -> Tuple[np.ndarray, Affine]:
    """
    Normalize ``Nx2`` points.

    Scale and translate such that mean is at 0 and mean distance from 0 is sqrt(2).

    :returns: Normalized array and affine matrix of the normalization.
    """
    assert pts.ndim == 2 and pts.shape[1] == 2
    assert out is None or out.shape == pts.shape

    _mean = pts.mean(axis=0)
    XX = np.subtract(pts, _mean, out=out)

    sx = (((XX**2).sum(axis=1) * 0.5) ** -0.5).mean()
    XX *= sx

    tx, ty = -_mean * sx
    A = Affine(sx, 0, tx, 0, sx, ty)

    return XX, A


def affine_from_pts(X: Sequence[XY[float]], Y: Sequence[XY[float]]) -> Affine:
    """
    Given points ``X,Y`` compute ``A``, such that: ``Y = A*X``.

    Needs at least 3 points.
    """

    assert len(X) == len(Y)
    assert len(X) >= 3

    n = len(X)

    YY = stack_xy(Y)
    XX = np.ones((n, 3), dtype="float64")
    for i, pt in enumerate(X):
        XX[i, :2] = pt.xy

    mm, *_ = np.linalg.lstsq(XX, YY, rcond=-1)
    a, d, b, e, c, f = mm.ravel()

    return Affine(a, b, c, d, e, f)


def resolution_from_affine(A: Affine) -> Resolution:
    """
    Compute resolution from Affine matrix.

    Deals with rotated case when needed.
    """
    if is_affine_st(A):
        rx, _, _, _, ry, *_ = A
        return resxy_(rx, ry)
    _, _, A_ = decompose_rws(A)
    rx, _, _, _, ry, *_ = A_
    return resxy_(rx, ry)


def edge_index(shape: SomeShape, closed: bool = False) -> Iterator[Tuple[int, int]]:
    """
    Like ``ndindex`` in numpy but for edge locations and limited to 2d.

    Returns a sequence of indexes along the edge of a 2d array of a given shape.
    Order is fixed, starting from ``(0, 0), (0, 1)...``. See example below

    .. code-block::

       # for shape=(10, 8)
       (0, 0), (0, 1) ... (0, 7),
       (1, 7), (2, 7) ... (9, 7),
       (9, 6), (9, 5) ... (9, 0),
       (8, 0), (7, 0) ... (1, 0),
       (0, 0) # Back to (0, 0) if closed=True

    """
    nx, ny = shape_(shape).xy
    ix, iy = 0, 0
    for ix in range(nx):
        yield (0, ix)
    for iy in range(1, ny):
        yield (iy, ix)
    for ix in range(ix - 1, -1, -1):
        yield (iy, ix)
    for iy in range(iy - 1, 0, -1):
        yield (iy, ix)
    if closed:
        yield (0, 0)


def quasi_random_r2(
    n: int,
    shape: Optional[SomeShape] = None,
    offset: int = 0,
) -> np.ndarray:
    """
    Generate quasi-random set of points.

    Returns quasi-random points in ``[0, 1)`` range. If ``shape=(ny, nx)`` is
    supplied, points are scaled to ``[0, nx), [0, ny)`` range.

    :param n: Number of points to generate
    :param offset: Generate from that offset in the sequence
    :returns: ``nx2`` numpy array

    References: http://extremelearning.com.au/unreasonable-effectiveness-of-quasirandom-sequences/
    """
    idx = np.arange(offset, offset + n, dtype="float32")
    aa = (0.7548776662466927, 0.5698402909980532)
    xy = np.fmod(np.outer(idx, aa), 1)

    if shape is not None:
        nx, ny = shape_(shape).wh
        xy[:, 0] *= nx
        xy[:, 1] *= ny

    return xy


class Bin1D:
    """
    Class for translating continous coordinates to bin index.

    Binning is defined using following parameters:

    :param sz: Bin size (positive floating point number)
    :param origin: Location of the left edge of bin ``0``
    :param direction: Direction of the bin index ``+1|-1``
    """

    __slots__ = ("sz", "origin", "direction")

    def __init__(self, sz: float, origin: float = 0.0, direction: Literal[1, -1] = 1):
        """
        Construct :py:class:`~odc.geo.math.Bin1D` object.

        :param sz:
           Size of each bin, must be positive
        :param origin:
           Location of the left edge of bin ``0``, defaults to ``0.0``
        :param direction:
           Default is to increment bin index left to right, supply ``-1`` to go the other way
        """
        assert direction in (-1, 1)
        assert sz > 0
        self.sz = sz
        self.origin = origin
        self.direction = direction

    def __getitem__(self, idx: int) -> Tuple[float, float]:
        """Convert index to interval."""
        _x = idx * self.sz * self.direction + self.origin
        return _x, _x + self.sz

    def bin(self, x: float) -> int:
        """Lookup bin index that ``x`` falls into."""
        ix = floor((x - self.origin) / self.sz)
        return int(self.direction * ix)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Bin1D):
            return False
        return (
            self.sz == other.sz
            and self.origin == other.origin
            and self.direction == other.direction
        )

    # pylint: disable=redefined-builtin
    @staticmethod
    def from_sample_bin(
        idx: int, bin: Tuple[float, float], direction: Literal[1, -1] = 1
    ) -> "Bin1D":
        """
        Construct :py:class:`~odc.geo.math.Bin1D` from a sample.

        :param idx:
           Index of a sample bin
        :param bin:
           ``x0, x1`` bin edges
        :param direction:
           Default is to increment bin index left to right, supply ``-1`` to go the other way
        """
        x0, x1 = bin
        assert x0 < x1
        sz = x1 - x0
        origin = x0 - sz * idx * direction

        return Bin1D(sz, origin, direction=direction)


class Poly2d:
    """
    Wrapper around polyval2d with input normalization.
    """

    def __init__(self, cc: np.ndarray, A: Affine) -> None:
        assert cc.shape in [(3, 3, 2), (2, 2, 2)]
        tol = 1e-6
        self._cc = cc
        self._A = A
        self._safe_to_grid = False

        sx, zx, tx, zy, sy, ty, *_ = A
        if abs(zx) < tol and abs(zy) < tol:
            self._norm = lambda x, y: (np.polyval([sx, tx], x), np.polyval([sy, ty], y))
            self._safe_to_grid = True
        else:
            self._norm = lambda x, y: A * (x, y)

    def __call__(self, x: Any, y: Any = None) -> Any:
        """
        Evaluate at points (x, y).

        When y is None, x is assumed to be Nx2 array.
        """
        if y is None:
            return self.__call__(x[..., 0], x[..., 1]).T
        x, y = self._norm(x, y)
        return polyval2d(x, y, self._cc)

    def grid2d(self, x: Any, y: Any) -> Any:
        """
        Evaluate on the Cartesian product of x and y.
        """
        assert self._safe_to_grid

        x, y = self._norm(x, y)
        return polygrid2d(x, y, self._cc)

    def with_input_transform(self, A: Affine) -> "Poly2d":
        """
        Make a new transform by chaining a linear mapping ``A`` on input side.

        usefull when cropping original image with GCPs.
        """
        return Poly2d(self._cc, self._A * A)

    @staticmethod
    def fit(aa: np.ndarray, bb: np.ndarray) -> "Poly2d":
        """
        Fit 2d polynomial that minimizes ``poly(aa) - bb``.

        Where ``aa``,``bb`` are point correspondences of ``Nx2`` shape.
        """
        assert aa.shape[1] == 2
        assert aa.shape == bb.shape
        N = aa.shape[0]
        if N < 3:
            raise ValueError(f"Need at least 3 points, got {N}")

        aa_, Ain = norm_xy(aa)
        bb_, Ab = norm_xy(bb)

        if N >= 9:
            return Poly2d._fit9(aa_, Ain, bb_, Ab)

        if N >= 4:
            return Poly2d._fit4(aa_, Ain, bb_, Ab)

        return Poly2d._fit3(aa_, Ain, bb_, Ab)

    @staticmethod
    def _fit9(aa: np.ndarray, Ain: Affine, bb: np.ndarray, Ab: Affine) -> "Poly2d":
        N = aa.shape[0]
        assert N >= 9
        x, y = aa.T
        AA = np.empty((N, 9), dtype="float64")

        #   1,     y,     y^2
        #   x,   x*y,   x*y^2
        # x^2, x^2*y, x^2*y^2

        AA[:, 0] = 1
        AA[:, 1] = y
        AA[:, 2] = y * y

        AA[:, 3] = x
        AA[:, 4] = x * y
        AA[:, 5] = AA[:, 4] * y

        AA[:, 6] = x * x
        AA[:, 7] = AA[:, 6] * y
        AA[:, 8] = AA[:, 7] * y

        cc, *_ = np.linalg.lstsq(AA, bb, rcond=-1)

        # denorm output side, assumes `sx==sy`
        s, _, tx, _, _, ty, *_ = ~Ab
        cc = cc * s
        cc[0, :2] += (tx, ty)

        cc = cc.reshape(3, 3, 2)

        return Poly2d(cc, Ain)

    @staticmethod
    def _fit4(aa: np.ndarray, Ain: Affine, bb: np.ndarray, Ab: Affine) -> "Poly2d":
        N = aa.shape[0]
        assert N >= 4
        x, y = aa.T
        AA = np.empty((N, 4), dtype="float64")

        #   1,     y
        #   x,   x*y

        AA[:, 0] = 1
        AA[:, 1] = y

        AA[:, 2] = x
        AA[:, 3] = x * y

        cc, *_ = np.linalg.lstsq(AA, bb, rcond=-1)
        assert cc.shape == (4, 2)

        # denorm output side, assumes `sx==sy`
        s, _, tx, _, _, ty, *_ = ~Ab
        cc = cc * s
        cc[0, :2] += (tx, ty)

        cc = cc.reshape(2, 2, 2)

        return Poly2d(cc, Ain)

    @staticmethod
    def _fit3(aa: np.ndarray, Ain: Affine, bb: np.ndarray, Ab: Affine) -> "Poly2d":
        N = aa.shape[0]
        assert N >= 3
        x, y = aa.T
        AA = np.empty((N, 3), dtype="float64")

        #   1,   y
        #   x,   _

        # 1, y, x
        AA[:, 0] = 1
        AA[:, 1] = y
        AA[:, 2] = x

        cc, *_ = np.linalg.lstsq(AA, bb, rcond=-1)
        assert cc.shape == (3, 2)

        # denorm output side, assumes `sx==sy`
        s, _, tx, _, _, ty, *_ = ~Ab
        cc = cc * s
        cc[0, :2] += (tx, ty)
        cc = np.vstack([cc, np.asarray([0, 0])]).reshape(2, 2, 2)

        return Poly2d(cc, Ain)


def extract_anchor(pix2wld: Affine, tol: float = 1e-3) -> AnchorEnum | XY[float]:
    def _anchor(px: float, tol: float) -> float:
        _, x = split_float(px)  # x in (-0.5, +0.5)
        x = (1 + x) if x < 0 else x  # x in [0, 1)
        x = snap_scale(maybe_zero(x, tol), tol)
        return 0 if x >= 1 else x

    anchor = tuple(_anchor(px, tol) for px in (~pix2wld) * (0, 0))
    if anchor == (0, 0):
        return AnchorEnum.EDGE
    if anchor == (0.5, 0.5):
        return AnchorEnum.CENTER
    return xy_(anchor)



import numpy as np
import pytest
from affine import Affine

from odc.geo import MaybeCRS, geom, wh_
from odc.geo.geobox import GeoBox, GeoboxTiles
from odc.geo.testutils import epsg3857


@pytest.mark.parametrize(
    "iso3, crs, resolution",
    [
        ("AUS", "epsg:4326", 0.1),
        ("AUS", "epsg:3577", 10_000),
        ("AUS", "epsg:3857", 10_000),
        ("NZL", "epsg:3857", 5_000),
    ],
)
def test_geoboxtiles_intersect(
    country: geom.Geometry, resolution: float, crs: MaybeCRS
):
    assert isinstance(country, geom.Geometry)
    assert country.crs == crs

    geobox = GeoBox.from_geopolygon(country, resolution=resolution, tight=True)
    assert geobox.crs == crs

    gbt = GeoboxTiles(geobox, (11, 13))
    assert gbt.base is geobox
    mm = gbt.grid_intersect(gbt)

    assert len(gbt.chunks) == 2
    assert (sum(gbt.chunks[0]), sum(gbt.chunks[1])) == gbt.base.shape

    for iy, ix in np.ndindex(gbt.shape.yx):
        idx = (iy, ix)
        assert idx in mm
        assert len(mm[idx]) == 1
        assert idx in mm[idx]

    gbt2 = GeoboxTiles(geobox.pad(2).to_crs(6933), (13, 11))
    mm = gbt.grid_intersect(gbt2)
    assert set(mm) == set(np.ndindex(gbt.shape.yx))

    gbt3 = GeoboxTiles(geobox.pad(2).rotate(1), (13, 11))
    mm = gbt.grid_intersect(gbt3)
    assert set(mm) == set(np.ndindex(gbt.shape.yx))


@pytest.mark.parametrize("use_chunks", [False, True])
def test_gbox_tiles(use_chunks):
    A = Affine.identity()
    H, W = (300, 200)
    h, w = (10, 20)
    gbox = GeoBox(wh_(W, H), A, epsg3857)
    tt = GeoboxTiles(gbox, (h, w))
    assert tt.shape == (300 / 10, 200 / 20)
    assert tt.base is gbox

    if use_chunks:
        tt = GeoboxTiles(gbox, tt.roi.chunks)

    # smoke test textual repr
    assert isinstance(str(tt), str)
    assert isinstance(repr(tt), str)

    # check ==
    assert tt == tt
    assert tt != "?"
    if use_chunks:
        assert tt == GeoboxTiles(tt.base, tt.chunks)
    else:
        assert tt == GeoboxTiles(tt.base, (h, w))

    assert tt[0, 0] == gbox[0:h, 0:w]
    assert tt[0, 1] == gbox[0:h, w : w + w]

    assert tt[4, 1].shape == (h, w)

    H, W = (11, 22)
    h, w = (10, 9)
    gbox = GeoBox(wh_(W, H), A, epsg3857)
    tt = GeoboxTiles(gbox, (h, w))
    assert tt.shape == (2, 3)
    assert tt[1, 2] == gbox[10:11, 18:22]

    # check .roi
    assert tt.base[tt.roi[1, 2]] == tt[1, 2]

    for idx in [tt.shape, (-33, 1)]:
        with pytest.raises(IndexError):
            _ = tt[idx]

        with pytest.raises(IndexError):
            tt.chunk_shape(idx)

    cc = np.zeros(tt.shape, dtype="int32")
    for idx in tt.tiles(gbox.extent):
        cc[idx] += 1
    np.testing.assert_array_equal(cc, np.ones(tt.shape))

    assert list(tt.tiles(gbox[:h, :w].extent)) == [(0, 0)]
    assert list(tt.tiles(gbox[:h, :w].extent.boundingbox)) == [(0, 0)]
    assert list(tt.tiles(gbox[:h, :w].extent.to_crs("epsg:4326"))) == [(0, 0)]

    (H, W) = (11, 22)
    (h, w) = (10, 20)
    tt = GeoboxTiles(GeoBox(wh_(W, H), A, epsg3857), (h, w))
    assert tt.chunk_shape((0, 0)) == (h, w)
    assert tt.chunk_shape((0, 1)) == (h, 2)
    assert tt.chunk_shape((1, 1)) == (1, 2)
    assert tt.chunk_shape((1, 0)) == (1, w)

    # check that overhang get's clamped properly
    assert tt.range_from_bbox(gbox.pad(2).boundingbox) == (
        range(0, tt.shape[0]),
        range(0, tt.shape[1]),
    )

    # bounding box in any projection should work
    assert tt.range_from_bbox(gbox.geographic_extent.boundingbox) == (
        range(0, tt.shape[0]),
        range(0, tt.shape[1]),
    )

    assert tt.range_from_bbox(tt[0, 0].extent.boundingbox) == (range(0, 1), range(0, 1))


@pytest.mark.parametrize("use_chunks", [False, True])
def test_gbox_tiles_roi(use_chunks):
    A = Affine.identity()
    H, W = (300, 200)
    h, w = (10, 20)
    gbox = GeoBox(wh_(W, H), A, epsg3857)
    tt = GeoboxTiles(gbox, (h, w))
    assert tt.shape == (H // h, W // w)
    assert tt.base is gbox

    if use_chunks:
        tt = GeoboxTiles(gbox, tt.roi.chunks)

    assert tt[:, :] == gbox
    assert tt.crop[:, :].base == gbox
    assert tt.crop[1, 2].base == tt[1, 2]

    assert tt.clip([(0, 0)]) == (tt.crop[0, 0], [(0, 0)])
    assert tt.clip([(1, 0)]) == (tt.crop[1, 0], [(0, 0)])
    assert tt.clip([(1, 2), (2, 5)]) == (
        tt.crop[1 : 2 + 1, 2 : 5 + 1],
        [(0, 0), (1, 3)],
    )

    for idx in np.ndindex(tt.shape.shape):
        bbox = tt.pix_bbox(idx)
        assert bbox.crs is None
        _idx = list(tt.tiles(bbox))
        assert len(_idx) == 1
        assert _idx[0] == idx



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from typing import Any, Iterable, Optional, Tuple, Union

import numpy as np
import rasterio.warp
from affine import Affine

from .gcp import GCPGeoBox
from .geobox import GeoBox
from .math import resolve_fill_value, resolve_nodata
from .types import Nodata, SomeNodata, wh_

# pylint: disable=invalid-name, too-many-arguments
Resampling = Union[str, int, rasterio.warp.Resampling]
_WRP_CRS = "epsg:3857"

__all__ = [
    "resampling_s2rio",
    "is_resampling_nn",
    "resolve_fill_value",
    "rio_warp_affine",
    "rio_reproject",
    "warp_affine",
]


def resampling_s2rio(name: str) -> rasterio.warp.Resampling:
    """
    Convert from string to rasterio.warp.Resampling enum, raises ValueError on bad input.
    """
    try:
        return getattr(rasterio.warp.Resampling, name.lower())
    except AttributeError:
        raise ValueError(f"Bad resampling parameter: {name}") from None


def is_resampling_nn(resampling: Resampling) -> bool:
    """
    :returns: True if resampling mode is nearest neighbour
    :returns: False otherwise
    """
    if isinstance(resampling, str):
        return resampling.lower() == "nearest"
    return resampling == rasterio.warp.Resampling.nearest


def rio_warp_affine(
    src: np.ndarray,
    dst: np.ndarray,
    A: Affine,
    resampling: Resampling,
    src_nodata: SomeNodata = "auto",
    dst_nodata: SomeNodata = "auto",
    **kwargs,
) -> np.ndarray:
    """
    Perform Affine warp using rasterio as backend library.

    :param        src: image as ndarray
    :param        dst: image as ndarray
    :param          A: Affine transform, maps from dst_coords to src_coords
    :param resampling: str|rasterio.warp.Resampling resampling strategy
    :param src_nodata: Value representing "no data" in the source image
    :param dst_nodata: Value to represent "no data" in the destination image

    :param     kwargs: any other args to pass to ``rasterio.warp.reproject``

    :returns: dst
    """
    assert src.ndim == dst.ndim
    assert src.ndim == 2
    sh, sw = src.shape
    dh, dw = dst.shape

    s_gbox = GeoBox(wh_(sw, sh), Affine.identity(), _WRP_CRS)
    d_gbox = GeoBox(wh_(dw, dh), A, _WRP_CRS)

    src_nodata = resolve_nodata(src_nodata, src.dtype)
    dst_nodata = resolve_nodata(dst_nodata, dst.dtype)
    fill_value = resolve_fill_value(dst_nodata, src_nodata, dst.dtype)

    return _rio_reproject(
        src, dst, s_gbox, d_gbox, resampling, src_nodata, fill_value, **kwargs
    )


# backward compatibility wrapper
warp_affine = rio_warp_affine


def rio_reproject(
    src: np.ndarray,
    dst: np.ndarray,
    s_gbox: Union[GeoBox, GCPGeoBox],
    d_gbox: GeoBox,
    resampling: Resampling,
    src_nodata: SomeNodata = "auto",
    dst_nodata: SomeNodata = "auto",
    ydim: Optional[int] = None,
    **kwargs,
) -> np.ndarray:
    """
    Perform reproject from ndarray->ndarray using rasterio as backend library.

    :param        src: image as ndarray
    :param        dst: image as ndarray
    :param     s_gbox: GeoBox of source image
    :param     d_gbox: GeoBox of destination image
    :param resampling: str|rasterio.warp.Resampling resampling strategy
    :param src_nodata: Value representing "no data" in the source image
    :param dst_nodata: Value to represent "no data" in the destination image
    :param       ydim: Which dimension is y-axis, next one must be x

    :param     kwargs: any other args to pass to ``rasterio.warp.reproject``

    :returns: dst
    """
    assert src.ndim == dst.ndim

    src_nodata = resolve_nodata(src_nodata, src.dtype)
    dst_nodata = resolve_nodata(dst_nodata, dst.dtype)
    fill_value = resolve_fill_value(dst_nodata, src_nodata, dst.dtype)

    if src.ndim == 2:
        return _rio_reproject(
            src,
            dst,
            s_gbox,
            d_gbox,
            resampling=resampling,
            src_nodata=src_nodata,
            dst_nodata=fill_value,
            **kwargs,
        )

    if ydim is None:
        # Assume last two dimensions are Y/X
        ydim = src.ndim - 2

    extra_dims = (*src.shape[:ydim], *src.shape[ydim + 2 :])
    # Selects each 2d plane in [...]YX[B] array
    slices: Iterable[Any] = (
        (*idx[:ydim], slice(None), slice(None), *idx[ydim:])
        for idx in np.ndindex(*extra_dims)
    )
    for roi in slices:
        _rio_reproject(
            src[roi],
            dst[roi],
            s_gbox,
            d_gbox,
            resampling=resampling,
            src_nodata=src_nodata,
            dst_nodata=fill_value,
            **kwargs,
        )
    return dst


def _rio_reproject(
    src: np.ndarray,
    dst: np.ndarray,
    s_gbox: Union[GCPGeoBox, GeoBox],
    d_gbox: GeoBox,
    resampling: Resampling,
    src_nodata: Nodata = None,
    dst_nodata: Nodata = None,
    **kwargs,
) -> np.ndarray:
    assert src.ndim == dst.ndim
    assert src.ndim == 2

    if "XSCALE" not in kwargs and "YSCALE" not in kwargs:
        # Work around for issue in GDAL
        #    https://github.com/OSGeo/gdal/issues/7750
        # Force GDAL into performing bilinear,cubic,lanczos with raidius=1px.
        # GDAL is trying to be smart and pick sampling radius based on scale
        # change, but does it without consideration for rotational component
        # of the transform, leading to blury output in some situations
        # See also:
        #    https://github.com/opendatacube/datacube-core/issues/1448
        kwargs.update(XSCALE=1, YSCALE=1)

    dtype_remap = {"int8": "int16", "bool": "uint8"}

    def _alias_or_convert(arr: np.ndarray) -> Tuple[np.ndarray, bool]:
        if arr.dtype.name not in dtype_remap:
            return arr, False
        wk_dtype = dtype_remap[arr.dtype.name]
        if arr.dtype.name == "bool":
            F, T = (np.array(v, dtype=wk_dtype) for v in [0, 255])
            return np.where(arr, T, F), True
        return arr.astype(wk_dtype), False

    if isinstance(resampling, str):
        resampling = resampling_s2rio(resampling)

    src_transform = None
    gcps = None

    if isinstance(s_gbox, GCPGeoBox):
        gcps = s_gbox.gcps()
    else:
        src_transform = s_gbox.transform

    # GDAL support for int8 is patchy, warp doesn't support it, so we need to convert to int16
    src, src_is_bool = _alias_or_convert(src)
    _dst, _ = _alias_or_convert(dst)

    rasterio.warp.reproject(
        src,
        _dst,
        src_transform=src_transform,
        gcps=gcps,
        src_crs=str(s_gbox.crs),
        dst_transform=d_gbox.transform,
        dst_crs=str(d_gbox.crs),
        resampling=resampling,
        src_nodata=src_nodata,
        dst_nodata=dst_nodata,
        **kwargs,
    )

    if dst is not _dst:
        # int8 workaround copy pixels back to int8
        if src_is_bool:
            # undo [0, 1] to [0, 255] stretching of the src
            np.copyto(dst, _dst > 127, casting="unsafe")
        else:
            np.copyto(dst, _dst, casting="unsafe")

    return dst



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Write Cloud Optimized GeoTIFFs from xarrays.
"""
from dataclasses import dataclass, field
from typing import Any, Dict, Iterator, Literal, Optional, Tuple, Union

import numpy as np

from ..geobox import GeoBox
from ..math import align_down_pow2, align_up
from ..types import Nodata, Shape2d, SomeShape, shape_, wh_

# pylint: disable=too-many-locals,too-many-branches,too-many-arguments,too-many-statements,too-many-instance-attributes

AxisOrder = Union[Literal["YX"], Literal["YXS"], Literal["SYX"]]

# map compressor name to level name in GDAL
GDAL_COMP: Dict[str, str] = {
    "DEFLATE": "ZLEVEL",
    "ADOBE_DEFLATE": "ZLEVEL",
    "ZSTD": "ZSTD_LEVEL",
    "WEBP": "WEBP_LEVEL",
    "LERC": "MAX_Z_ERROR",
    "LERC_DEFLATE": "MAX_Z_ERROR",
    "LERC_ZSTD": "MAX_Z_ERROR",
    "JPEG": "JPEG_QUALITY",
}

GEOTIFF_TAGS = {
    34264,  # ModelTransformation
    34735,  # GeoKeyDirectory
    34736,  # GeoDoubleParams
    34737,  # GeoAsciiParams
    33550,  # ModelPixelScale
    33922,  # ModelTiePoint
    #
    42112,  # GDAL_METADATA
    42113,  # GDAL_NODATA
    #
    # probably never used in the wild
    33920,  # IrasB Transformation Matrix
    50844,  # RPCCoefficientTag
}


@dataclass
class CogMeta:
    """
    COG metadata.
    """

    axis: AxisOrder
    shape: Shape2d
    tile: Shape2d
    nsamples: int
    dtype: Any
    compression: int
    predictor: int
    compressionargs: Dict[str, Any] = field(default_factory=dict, repr=False)
    gbox: Optional[GeoBox] = None
    overviews: Tuple["CogMeta", ...] = field(default=(), repr=False)
    nodata: Nodata = None

    def _pix_shape(self, shape: Shape2d) -> Tuple[int, ...]:
        if self.axis == "YX":
            return shape.shape
        if self.axis == "YXS":
            return (*shape.shape, self.nsamples)
        return (self.nsamples, *shape.shape)

    @property
    def chunks(self) -> Tuple[int, ...]:
        return self._pix_shape(self.tile)

    @property
    def pix_shape(self) -> Tuple[int, ...]:
        return self._pix_shape(self.shape)

    @property
    def num_planes(self):
        if self.axis == "SYX":
            return self.nsamples
        return 1

    def flatten(self) -> Tuple["CogMeta", ...]:
        return (self, *self.overviews)

    @property
    def chunked(self) -> Shape2d:
        """
        Shape in chunks.
        """
        ny, nx = ((N + n - 1) // n for N, n in zip(self.shape.yx, self.tile.yx))
        return shape_((ny, nx))

    @property
    def num_tiles(self):
        ny, nx = self.chunked.yx
        return self.num_planes * ny * nx

    def tidx(self, sample_idx: Optional[int] = None) -> Iterator[Tuple[int, int, int]]:
        """``[([sample|plane]_idx, iy, ix), ...]``"""
        if sample_idx is not None:
            assert sample_idx < self.num_planes
            yield from ((sample_idx, y, x) for y, x in np.ndindex(self.chunked.yx))
        else:
            yield from np.ndindex((self.num_planes, *self.chunked.yx))

    def flat_tile_idx(self, idx: Tuple[int, int, int]) -> int:
        """Convert from sample,iy,ix to flat tile index."""
        ns = self.num_planes
        ny, nx = self.chunked.yx
        for n, i in zip((ns, ny, nx), idx):
            if i < 0 or i >= n:
                raise IndexError()

        sample, y, x = idx
        return sample * (ny * nx) + y * nx + x

    def cog_tidx(self) -> Iterator[Tuple[int, int, int, int]]:
        """``[(ifd_idx, plane_idx, iy, ix), ...]``"""
        idx_layers = list(enumerate(self.flatten()))[::-1]
        for idx, mm in idx_layers:
            yield from ((idx, pi, yi, xi) for pi, yi, xi in mm.tidx())

    def __dask_tokenize__(self):
        return (
            "odc.CogMeta",
            self.axis,
            *self.shape.yx,
            *self.tile.yx,
            self.nsamples,
            self.dtype,
            self.compression,
            self.predictor,
            self.gbox,
            len(self.overviews),
            self.nodata,
        )


def adjust_blocksize(block: int, dim: int = 0) -> int:
    if 0 < dim < block:
        return align_up(dim, 16)
    return align_up(block, 16)


def norm_blocksize(block: Union[int, Tuple[int, int]]) -> Tuple[int, int]:
    if isinstance(block, int):
        block = adjust_blocksize(block)
        return (block, block)

    b1, b2 = map(adjust_blocksize, block)
    return (b1, b2)


def num_overviews(block: int, dim: int) -> int:
    c = 0
    while block < dim:
        dim = dim // 2
        c += 1
    return c


def compute_cog_spec(
    data_shape: SomeShape,
    tile_shape: SomeShape,
    *,
    max_pad: Optional[int] = None,
) -> Tuple[Shape2d, Shape2d, int]:
    data_shape = shape_(data_shape)
    tile_shape = shape_(shape_(tile_shape).map(adjust_blocksize))
    n1, n2 = (num_overviews(b, dim) for dim, b in zip(data_shape.xy, tile_shape.xy))
    n = max(n1, n2)
    pad = 2**n
    if max_pad is not None and max_pad < pad:
        pad = 0 if max_pad == 0 else align_down_pow2(max_pad)

    if pad > 0:
        data_shape = shape_(data_shape.map(lambda d: align_up(d, pad)))
    return (data_shape, tile_shape, n)


def cog_gbox(
    gbox: GeoBox,
    *,
    tile: Union[None, int, Tuple[int, int], Shape2d] = None,
    nlevels: Optional[int] = None,
) -> GeoBox:
    """
    Return padded gbox with safe dimensions for COG.

    1. Compute number of desired overviews
    2. Expand gebox on the right/bottom to have exact pixel shrink across all levels
    """

    if nlevels is None:
        if tile is None:
            tile = wh_(256, 256)
        if isinstance(tile, int):
            tile = wh_(tile, tile)
        new_shape, _, _ = compute_cog_spec(gbox.shape, tile)
    else:
        pad = 1 << nlevels
        new_shape = shape_(gbox.shape.map(lambda d: align_up(d, pad)))
    return gbox.expand(new_shape)


def yaxis_from_shape(
    shape: Tuple[int, ...], gbox: Optional[GeoBox] = None
) -> Tuple[AxisOrder, int]:
    ndim = len(shape)

    if ndim == 2:
        return "YX", 0

    if ndim != 3:
        raise ValueError("Can only work with 2-d or 3-d data")
    if shape[-1] in (3, 4):  # YXS in RGB(A)
        return "YXS", 0

    if gbox is None:
        return "SYX", 1
    if gbox.shape == shape[:2]:  # YXS
        return "YXS", 0
    if gbox.shape == shape[1:]:  # SYX
        return "SYX", 1

    raise ValueError("Geobox and image shape do not match")



from ._rio import to_cog, write_cog, write_cog_layers
from ._shared import CogMeta, cog_gbox
from ._tifffile import save_cog_with_dask

__all__ = [
    "CogMeta",
    "cog_gbox",
    "to_cog",
    "write_cog",
    "write_cog_layers",
    "save_cog_with_dask",
]
